# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for spherical harmonics."""

# pylint: disable=line-too-long
# pylint: disable=protected-access

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys

from absl.testing import parameterized
import numpy as np
import tensorflow as tf

from tensorflow_graphics.math import math_helpers
from tensorflow_graphics.math import spherical_harmonics
from tensorflow_graphics.util import asserts
from tensorflow_graphics.util import test_case


class SphericalHarmonicsTest(test_case.TestCase):

  def test_evaluate_legendre_polynomial_preset(self):
    """Tests that evaluate_legendre_polynomial generates expected results."""
    tensor_size = np.random.randint(1, 3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    x = np.random.uniform(size=tensor_shape)

    with self.subTest(name="l_0_m_0"):
      l = tf.constant(0, shape=tensor_shape)
      m = tf.constant(0, shape=tensor_shape)
      pred = spherical_harmonics.evaluate_legendre_polynomial(l, m, x)
      gt = np.ones_like(x)

      self.assertAllClose(pred, gt)

    with self.subTest(name="l_1_m_0"):
      l = tf.constant(1, shape=tensor_shape)
      m = tf.constant(0, shape=tensor_shape)
      pred = spherical_harmonics.evaluate_legendre_polynomial(l, m, x)
      gt = x

      self.assertAllClose(pred, gt)

    with self.subTest(name="l_1_m_1"):
      l = tf.constant(1, shape=tensor_shape)
      m = tf.constant(1, shape=tensor_shape)
      pred = spherical_harmonics.evaluate_legendre_polynomial(l, m, x)
      gt = -tf.sqrt(1.0 - x * x)

      self.assertAllClose(pred, gt)

    with self.subTest(name="l_2_m_0"):
      l = tf.constant(2, shape=tensor_shape)
      m = tf.constant(0, shape=tensor_shape)
      pred = spherical_harmonics.evaluate_legendre_polynomial(l, m, x)
      gt = 0.5 * (3.0 * x * x - 1.0)

      self.assertAllClose(pred, gt)

    with self.subTest(name="l_2_m_1"):
      l = tf.constant(2, shape=tensor_shape)
      m = tf.constant(1, shape=tensor_shape)
      pred = spherical_harmonics.evaluate_legendre_polynomial(l, m, x)
      gt = -3.0 * x * tf.sqrt(1.0 - x * x)

      self.assertAllClose(pred, gt)

    with self.subTest(name="l_2_m_2"):
      l = tf.constant(2, shape=tensor_shape)
      m = tf.constant(2, shape=tensor_shape)
      pred = spherical_harmonics.evaluate_legendre_polynomial(l, m, x)
      gt = 3.0 * (1.0 - x * x)

      self.assertAllClose(pred, gt)

    with self.subTest(name="l_3_m_0"):
      l = tf.constant(3, shape=tensor_shape)
      m = tf.constant(0, shape=tensor_shape)
      pred = spherical_harmonics.evaluate_legendre_polynomial(l, m, x)
      gt = 0.5 * x * (5.0 * x * x - 3.0)

      self.assertAllClose(pred, gt)

    with self.subTest(name="l_3_m_1"):
      l = tf.constant(3, shape=tensor_shape)
      m = tf.constant(1, shape=tensor_shape)
      pred = spherical_harmonics.evaluate_legendre_polynomial(l, m, x)
      gt = 1.5 * (1.0 - 5.0 * x * x) * tf.sqrt(1.0 - x * x)

      self.assertAllClose(pred, gt)

    with self.subTest(name="l_3_m_2"):
      l = tf.constant(3, shape=tensor_shape)
      m = tf.constant(2, shape=tensor_shape)
      pred = spherical_harmonics.evaluate_legendre_polynomial(l, m, x)
      gt = 15.0 * x * (1.0 - x * x)

      self.assertAllClose(pred, gt)

    with self.subTest(name="l_3_m_3"):
      l = tf.constant(3, shape=tensor_shape)
      m = tf.constant(3, shape=tensor_shape)
      pred = spherical_harmonics.evaluate_legendre_polynomial(l, m, x)
      gt = -15.0 * tf.pow(1.0 - x * x, 1.5)

      self.assertAllClose(pred, gt)

  @parameterized.parameters(
      ("Not all batch dimensions are broadcast-compatible", tf.int32, tf.int32,
       (3, 3), (3, 3), (2, 3)),
      ("Not all batch dimensions are broadcast-compatible", tf.int32, tf.int32,
       (None, 5, 2), (None, 1, 2), (3, 1)),
      ("must be of an integer type", tf.float32, tf.int32, (3,), (3,), (1,)),
      ("must be of an integer type", tf.int32, tf.float32, (3,), (3,), (1,)),
  )
  def test_evaluate_legendre_polynomial_raised(self, error_msg, degree_dtype,
                                               order_dtype, *shapes):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(
        spherical_harmonics.evaluate_legendre_polynomial,
        error_msg,
        shapes,
        dtypes=(degree_dtype, order_dtype, tf.float32))

  def test_evaluate_legendre_polynomial_exceptions_l_raised(self):
    """Tests that an exception is raised when l is not in the expected range."""
    l = np.random.randint(-10, -1, size=(1,))
    m = np.random.randint(5, 10, size=(1,))
    x = np.random.uniform(size=(1,)) * 2.0

    with self.assertRaises(tf.errors.InvalidArgumentError):
      self.evaluate(spherical_harmonics.evaluate_legendre_polynomial(l, m, x))

  def test_evaluate_legendre_polynomial_exceptions_m_raised(self):
    """Tests that an exception is raised when m is not in the expected range."""
    l = np.random.randint(1, 4, size=(1,))
    m = np.random.randint(5, 10, size=(1,))
    x = np.random.uniform(size=(1,)) * 2.0

    with self.assertRaises(tf.errors.InvalidArgumentError):
      self.evaluate(spherical_harmonics.evaluate_legendre_polynomial(l, m, x))

  def test_evaluate_legendre_polynomial_exceptions_x_raised(self):
    """Tests that an exception is raised when x is not in the expected range."""
    l = np.random.randint(4, 10, size=(1,))
    m = np.random.randint(0, 4, size=(1,))
    x = np.random.uniform(
        1.0 + asserts.select_eps_for_addition(tf.float32), 2.0, size=(1,))

    with self.assertRaises(tf.errors.InvalidArgumentError):
      self.evaluate(spherical_harmonics.evaluate_legendre_polynomial(l, m, x))

  def test_evaluate_legendre_polynomial_jacobian_random(self):
    """Tests the Jacobian of evaluate_legendre_polynomial."""
    tensor_size = np.random.randint(1, 3)
    tensor_shape = np.random.randint(1, 5, size=(tensor_size)).tolist()
    l_init = np.random.randint(5, 10, size=tensor_shape)
    m_init = np.random.randint(0, 4, size=tensor_shape)
    x_init = np.random.uniform(-1.0, 1.0, size=tensor_shape)

    def evaluate_legendre_polynomial_fn(x):
      return spherical_harmonics.evaluate_legendre_polynomial(l_init, m_init, x)

    self.assert_jacobian_is_correct_fn(
        evaluate_legendre_polynomial_fn, [x_init], atol=1e-3)

  @parameterized.parameters(
      ((4,), (4,)),
      ((None, 9), (None, 9)),
      ((None, 3, 9), (None, 1, 9)),
  )
  def test_integration_product_not_raised(self, *shape):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(spherical_harmonics.integration_product,
                                        shape)

  @parameterized.parameters(
      ("must have the same number of dimensions in axes", (1,), (2,)),)
  def test_integration_product_raised(self, error_msg, *shape):
    """Tests that the shape exception is raised."""
    self.assert_exception_is_raised(spherical_harmonics.integration_product,
                                    error_msg, shape)

  def test_generate_l_m_permutations_preset(self):
    """Tests that generate_l_m_permutations produces the expected results."""
    l, m = spherical_harmonics.generate_l_m_permutations(2)

    self.assertAllEqual(l, (0, 1, 1, 1, 2, 2, 2, 2, 2))
    self.assertAllEqual(m, (0, -1, 0, 1, -2, -1, 0, 1, 2))

  def test_generate_l_m_zonal_preset(self):
    """Tests that generate_l_m_zonal produces the expected results."""
    l, m = spherical_harmonics.generate_l_m_zonal(2)

    self.assertAllEqual(l, (0, 1, 2))
    self.assertAllEqual(m, (0, 0, 0))

  @parameterized.parameters(
      (((2, -1)), (0.25752,)),
      (((8, 6)), (5.5710e-06,)),
      (((8, -6)), (5.5710e-06,)),
  )
  def test_spherical_harmonics_normalization_preset(self, test_inputs,
                                                    test_outputs):
    self.assert_output_is_correct(
        spherical_harmonics._spherical_harmonics_normalization, test_inputs,
        test_outputs)

  @parameterized.parameters(
      ("must have the same number of dimensions in axes", tf.int32, tf.int32,
       (3,), (2,), (1,), (1,)),
      ("must have exactly 1 dimensions in axis -1", tf.int32, tf.int32, (3,),
       (3,), (1,), (2,)),
      ("must have exactly 1 dimensions in axis -1", tf.int32, tf.int32, (3,),
       (3,), (2, 2), (2, 2)),
      ("Not all batch dimensions are identical.", tf.int32, tf.int32, (3, 3),
       (3, 3), (2, 1), (2, 1)),
      ("Not all batch dimensions are identical.", tf.int32, tf.int32,
       (None, 5, 2), (None, 1, 2), (5, 1), (1, 1)),
      ("must be of an integer type", tf.float32, tf.int32, (3,), (3,), (1,),
       (1,)),
      ("must be of an integer type", tf.int32, tf.float32, (3,), (3,), (1,),
       (1,)),
  )
  def test_evaluate_spherical_harmonics_raised(self, error_msg, degree_dtype,
                                               order_dtype, *shapes):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(
        spherical_harmonics.evaluate_spherical_harmonics,
        error_msg,
        shapes,
        dtypes=(degree_dtype, order_dtype, tf.float32, tf.float32))

  def test_evaluate_spherical_harmonics_exception_l_raised(self):
    """Tests that an exception is raised when l is not in the expected range."""
    l = np.random.randint(-10, -1, size=(1,))
    m = np.random.randint(5, 10, size=(1,))
    theta = np.random.uniform(0.0, np.pi, size=(1,))
    phi = np.random.uniform(0.0, 2.0 * np.pi, size=(1,))

    with self.assertRaises(tf.errors.InvalidArgumentError):
      self.evaluate(
          spherical_harmonics.evaluate_spherical_harmonics(l, m, theta, phi))

  def test_evaluate_spherical_harmonics_exception_m_raised(self):
    """Tests that an exception is raised when m is not in the expected range."""
    l = np.random.randint(1, 4, size=(1,))
    m = np.random.randint(5, 10, size=(1,))
    theta = np.random.uniform(0.0, np.pi, size=(1,))
    phi = np.random.uniform(0.0, 2.0 * np.pi, size=(1,))

    with self.assertRaises(tf.errors.InvalidArgumentError):
      self.evaluate(
          spherical_harmonics.evaluate_spherical_harmonics(l, m, theta, phi))

    m = np.random.randint(-10, -5, size=(1,))

    with self.assertRaises(tf.errors.InvalidArgumentError):
      self.evaluate(
          spherical_harmonics.evaluate_spherical_harmonics(l, m, theta, phi))

  def test_evaluate_spherical_harmonics_exception_theta_raised(self):
    """Tests exceptions on the values of theta."""
    l = np.random.randint(1, 4, size=(1,))
    m = np.random.randint(5, 10, size=(1,))
    theta = np.random.uniform(
        np.pi + sys.float_info.epsilon, 2.0 * np.pi, size=(1,))
    phi = np.random.uniform(0.0, 2.0 * np.pi, size=(1,))

    with self.assertRaises(tf.errors.InvalidArgumentError):
      self.evaluate(
          spherical_harmonics.evaluate_spherical_harmonics(l, m, theta, phi))

    theta = np.random.uniform(-np.pi, 0.0 - sys.float_info.epsilon, size=(1,))

    with self.assertRaises(tf.errors.InvalidArgumentError):
      self.evaluate(
          spherical_harmonics.evaluate_spherical_harmonics(l, m, theta, phi))

  def test_evaluate_spherical_harmonics_exception_phi_raised(self):
    """Tests exceptions on the values of phi."""
    l = np.random.randint(1, 4, size=(1,))
    m = np.random.randint(5, 10, size=(1,))
    theta = np.random.uniform(0.0, np.pi, size=(1,))
    phi = np.random.uniform(
        2.0 * np.pi + sys.float_info.epsilon, 4.0 * np.pi, size=(1,))

    with self.assertRaises(tf.errors.InvalidArgumentError):
      self.evaluate(
          spherical_harmonics.evaluate_spherical_harmonics(l, m, theta, phi))

    phi = np.random.uniform(
        -2.0 * np.pi, 0.0 - sys.float_info.epsilon, size=(1,))

    with self.assertRaises(tf.errors.InvalidArgumentError):
      self.evaluate(
          spherical_harmonics.evaluate_spherical_harmonics(l, m, theta, phi))

  @parameterized.parameters(
      ((3,), (3,), (1,), (1,)),
      ((5, 2), (5, 2), (5, 1), (5, 1)),
  )
  def test_evaluate_spherical_harmonics_not_raised(self, *shape):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(
        spherical_harmonics.evaluate_spherical_harmonics,
        shape,
        dtypes=(tf.int32, tf.int32, tf.float32, tf.float32))

  def test_evaluate_spherical_harmonics_preset(self):
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    pt_3d = tf.convert_to_tensor(
        value=np.random.uniform(size=tensor_shape + [3]))
    pt_3d = tf.math.l2_normalize(pt_3d, axis=-1)
    x, y, z = tf.unstack(pt_3d, axis=-1)
    x = tf.expand_dims(x, axis=-1)
    y = tf.expand_dims(y, axis=-1)
    z = tf.expand_dims(z, axis=-1)
    pt_spherical = math_helpers.cartesian_to_spherical_coordinates(pt_3d)
    _, theta, phi = tf.unstack(pt_spherical, axis=-1)
    theta = tf.expand_dims(theta, axis=-1)
    phi = tf.expand_dims(phi, axis=-1)
    ones = tf.ones_like(z)

    with self.subTest(name="l_0_m_0"):
      l = tf.zeros_like(theta, dtype=tf.int32)
      m = tf.zeros_like(theta, dtype=tf.int32)
      val = spherical_harmonics.evaluate_spherical_harmonics(l, m, theta, phi)
      gt = ones / (2.0 * tf.sqrt(ones * np.pi))

      self.assertAllClose(val, gt)

    with self.subTest(name="l_1_m_-1"):
      l = tf.ones_like(theta, dtype=tf.int32)
      m = -tf.ones_like(theta, dtype=tf.int32)
      val = spherical_harmonics.evaluate_spherical_harmonics(l, m, theta, phi)
      gt = -tf.sqrt(3.0 * ones) * y / (2.0 * tf.sqrt(ones * np.pi))

      self.assertAllClose(val, gt)

    with self.subTest(name="l_1_m_0"):
      l = tf.ones_like(theta, dtype=tf.int32)
      m = tf.zeros_like(theta, dtype=tf.int32)
      val = spherical_harmonics.evaluate_spherical_harmonics(l, m, theta, phi)
      gt = (tf.sqrt(3.0 * ones) * z) / (2.0 * tf.sqrt(ones * np.pi))

      self.assertAllClose(val, gt)

    with self.subTest(name="l_1_m_1"):
      l = tf.ones_like(theta, dtype=tf.int32)
      m = tf.ones_like(theta, dtype=tf.int32)
      val = spherical_harmonics.evaluate_spherical_harmonics(l, m, theta, phi)
      gt = -tf.sqrt(3.0 * ones) * x / (2.0 * tf.sqrt(ones * np.pi))

      self.assertAllClose(val, gt)

    with self.subTest(name="l_2_m-2"):
      l = 2 * tf.ones_like(theta, dtype=tf.int32)
      m = -2 * tf.ones_like(theta, dtype=tf.int32)
      val = spherical_harmonics.evaluate_spherical_harmonics(l, m, theta, phi)
      gt = tf.sqrt(15.0 * ones) * y * x / (2.0 * tf.sqrt(ones * np.pi))

      self.assertAllClose(val, gt)

    with self.subTest(name="l_2_m_-1"):
      l = 2 * tf.ones_like(theta, dtype=tf.int32)
      m = -tf.ones_like(theta, dtype=tf.int32)
      val = spherical_harmonics.evaluate_spherical_harmonics(l, m, theta, phi)
      gt = -tf.sqrt(15.0 * ones) * y * z / (2.0 * tf.sqrt(ones * np.pi))

      self.assertAllClose(val, gt)

    with self.subTest(name="l_2_m_0"):
      l = 2 * tf.ones_like(theta, dtype=tf.int32)
      m = tf.zeros_like(theta, dtype=tf.int32)
      val = spherical_harmonics.evaluate_spherical_harmonics(l, m, theta, phi)
      gt = tf.sqrt(
          5.0 * ones) * (3.0 * z * z - ones) / (4.0 * tf.sqrt(ones * np.pi))

      self.assertAllClose(val, gt)

    with self.subTest(name="l_2_m_1"):
      l = 2 * tf.ones_like(theta, dtype=tf.int32)
      m = tf.ones_like(theta, dtype=tf.int32)
      val = spherical_harmonics.evaluate_spherical_harmonics(l, m, theta, phi)
      gt = -tf.sqrt(15.0 * ones) * x * z / (2.0 * tf.sqrt(ones * np.pi))

      self.assertAllClose(val, gt)

    with self.subTest(name="l_2_m_2"):
      l = 2 * tf.ones_like(theta, dtype=tf.int32)
      m = 2 * tf.ones_like(theta, dtype=tf.int32)
      val = spherical_harmonics.evaluate_spherical_harmonics(l, m, theta, phi)
      gt = tf.sqrt(
          15.0 * ones) * (x * x - y * y) / (4.0 * tf.sqrt(ones * np.pi))

      self.assertAllClose(val, gt)

  def test_evaluate_spherical_harmonics_jacobian_random(self):
    """Test the Jacobian of the evaluate_spherical_harmonics function."""
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 5, size=(tensor_size)).tolist()
    l_init = np.random.randint(5, 10, size=tensor_shape)
    l_init = np.expand_dims(l_init, axis=-1)
    m_init = np.random.randint(0, 4, size=tensor_shape)
    m_init = np.expand_dims(m_init, axis=-1)
    theta_init = np.random.uniform(0.0, np.pi, size=tensor_shape)
    theta_init = np.expand_dims(theta_init, axis=-1)
    phi_init = np.random.uniform(0.0, 2.0 * np.pi, size=tensor_shape)
    phi_init = np.expand_dims(phi_init, axis=-1)

    def evaluate_spherical_harmonics_fn(theta, phi):
      return spherical_harmonics.evaluate_spherical_harmonics(
          l_init, m_init, theta, phi)

    self.assert_jacobian_is_correct_fn(evaluate_spherical_harmonics_fn,
                                       [theta_init, phi_init])

  def test_rotate_zonal_harmonics_jacobian_random(self):
    """Tests the jacobian of rotate_zonal_harmonics."""
    dtype = tf.float64
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 5, size=(tensor_size)).tolist()
    theta_init = np.random.uniform(0.0, np.pi, size=tensor_shape + [1])
    phi_init = np.random.uniform(0.0, 2.0 * np.pi, size=tensor_shape + [1])
    zonal_coeffs = tf.convert_to_tensor(
        value=np.random.uniform(-1.0, 1.0, size=[3]), dtype=dtype)

    def rotate_zonal_harmonics_fn(theta, phi):
      return spherical_harmonics.rotate_zonal_harmonics(zonal_coeffs, theta,
                                                        phi)

    self.assert_jacobian_is_correct_fn(rotate_zonal_harmonics_fn,
                                       [theta_init, phi_init])

  @parameterized.parameters(
      ((4,), (3, 1), (3, 1)),
      ((4,), (3, 3, 1), (3, 3, 1)),
  )
  def test_rotate_zonal_harmonics_not_raised(self, *shape):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(
        spherical_harmonics.rotate_zonal_harmonics, shape)

  @parameterized.parameters(
      ("must have exactly 1 dimensions in axis -1", (4,), (5, 4), (5, 1)),
      ("must have a rank of 1, but it has rank", (4, 1), (5, 1), (5, 1)),
      ("must have exactly 1 dimensions in axis -1", (4,), (5, 2), (5, 2)),
  )
  def test_rotate_zonal_harmonics_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(spherical_harmonics.rotate_zonal_harmonics,
                                    error_msg, shapes)

  def test_rotate_zonal_harmonics_random(self):
    """Tests the outputs of test_rotate_zonal_harmonics."""
    dtype = tf.float64
    max_band = 2
    zonal_coeffs = tf.constant(
        np.random.uniform(-1.0, 1.0, size=[3]), dtype=dtype)
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    theta = tf.constant(
        np.random.uniform(0.0, np.pi, size=tensor_shape + [1]), dtype=dtype)
    phi = tf.constant(
        np.random.uniform(0.0, 2.0 * np.pi, size=tensor_shape + [1]),
        dtype=dtype)

    rotated_zonal_coeffs = spherical_harmonics.rotate_zonal_harmonics(
        zonal_coeffs, theta, phi)
    zonal_coeffs = spherical_harmonics.tile_zonal_coefficients(zonal_coeffs)
    l, m = spherical_harmonics.generate_l_m_permutations(max_band)
    l = tf.broadcast_to(l, tensor_shape + l.shape.as_list())
    m = tf.broadcast_to(m, tensor_shape + m.shape.as_list())
    theta_zero = tf.constant(0.0, shape=tensor_shape + [1], dtype=dtype)
    phi_zero = tf.constant(0.0, shape=tensor_shape + [1], dtype=dtype)
    gt = zonal_coeffs * spherical_harmonics.evaluate_spherical_harmonics(
        l, m, theta_zero, phi_zero)
    gt = tf.reduce_sum(input_tensor=gt, axis=-1)
    pred = rotated_zonal_coeffs * spherical_harmonics.evaluate_spherical_harmonics(
        l, m, theta + theta_zero, phi + phi_zero)
    pred = tf.reduce_sum(input_tensor=pred, axis=-1)

    self.assertAllClose(gt, pred)

  def test_tile_zonal_coefficients_jacobian_random(self):
    """Tests the jacobian of tile_zonal_coefficients."""
    zonal_coeffs_init = np.random.uniform(size=(1,))

    self.assert_jacobian_is_correct_fn(
        spherical_harmonics.tile_zonal_coefficients, [zonal_coeffs_init])

  @parameterized.parameters(
      ((2,)),)
  def test_tile_zonal_coefficients_not_raised(self, *shape):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(
        spherical_harmonics.tile_zonal_coefficients, shape)

  @parameterized.parameters(
      ("must have a rank of 1, but it has rank", (2, 2)),)
  def test_tile_zonal_coefficients_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(spherical_harmonics.tile_zonal_coefficients,
                                    error_msg, shapes)

  def test_tile_zonal_coefficients_preset(self):
    """Tests that tile_zonal_coefficients produces the expected results."""
    self.assertAllEqual(
        spherical_harmonics.tile_zonal_coefficients((0, 1, 2)),
        (0, 1, 1, 1, 2, 2, 2, 2, 2))


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tensorflow point utility functions."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from tensorflow_graphics.math import vector
from tensorflow_graphics.util import asserts
from tensorflow_graphics.util import export_api
from tensorflow_graphics.util import shape


def distance_to_ray(point, origin, direction, keepdims=True, name=None):
  """Computes the distance from a M-d point to a M-d ray.

  Note:
    In the following, A1 to An are optional batch dimensions, which must be
    broadcast compatible.

  Args:
    point: A tensor of shape `[A1, ..., An, M]`.
    origin: A tensor of shape `[A1, ..., An, M]`.
    direction: A tensor of shape `[A1, ..., An, M]`. The last dimension must be
      normalized.
    keepdims: A `bool`, whether to keep the last dimension with length 1 or to
      remove it.
    name: A name for this op. Defaults to "point_distance_to_ray".

  Returns:
    A tensor of shape `[A1, ..., An, 1]` containing the distance from each point
    to the corresponding ray.

  Raises:
    ValueError: If the shape of `point`, `origin`, or 'direction' is not
    supported.
  """
  with tf.compat.v1.name_scope(name, "point_distance_to_ray",
                               [point, origin, direction]):
    point = tf.convert_to_tensor(value=point)
    origin = tf.convert_to_tensor(value=origin)
    direction = tf.convert_to_tensor(value=direction)

    shape.compare_dimensions((point, origin, direction), -1,
                             ("point", "origin", "direction"))
    shape.compare_batch_dimensions(
        tensors=(point, origin, direction),
        last_axes=-2,
        broadcast_compatible=True)
    direction = asserts.assert_normalized(direction)

    vec = point - origin
    dot = vector.dot(vec, direction)
    vec -= dot * direction
    return tf.norm(tensor=vec, axis=-1, keepdims=keepdims)


def project_to_ray(point, origin, direction, name=None):
  """Computes the projection of a M-d point on a M-d ray.

  Note:
    In the following, A1 to An are optional batch dimensions, which must be
    broadcast compatible.

  Args:
    point: A tensor of shape `[A1, ..., An, M]`.
    origin: A tensor of shape `[A1, ..., An, M]`.
    direction: A tensor of shape `[A1, ..., An, M]`. The last dimension must be
      normalized.
    name: A name for this op. Defaults to "point_project_to_ray".

  Returns:
    A tensor of shape `[A1, ..., An, M]` containing the projected point.

  Raises:
    ValueError: If the shape of `point`, `origin`, or 'direction' is not
    supported.
  """
  with tf.compat.v1.name_scope(name, "point_project_to_ray",
                               [point, origin, direction]):
    point = tf.convert_to_tensor(value=point)
    origin = tf.convert_to_tensor(value=origin)
    direction = tf.convert_to_tensor(value=direction)

    shape.compare_dimensions((point, origin, direction), -1,
                             ("point", "origin", "direction"))
    shape.compare_batch_dimensions(
        tensors=(point, origin, direction),
        last_axes=-2,
        broadcast_compatible=True)
    direction = asserts.assert_normalized(direction)

    vec = point - origin
    dot = vector.dot(vec, direction)
    return origin + dot * direction


# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module implements various sparse data utilities for graphs and meshes."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from tensorflow_graphics.util import shape


def _is_dynamic_shape(tensors):
  """Helper function to test if any tensor in a list has a dynamic shape.

  Args:
    tensors: A list or tuple of tensors with shapes to test.

  Returns:
    True if any tensor in the list has a dynamic shape, False otherwise.
  """
  if not isinstance(tensors, (list, tuple)):
    raise ValueError("'tensors' must be list of tuple.")
  return not all([shape.is_static(tensor.shape) for tensor in tensors])


def check_valid_graph_convolution_input(data, neighbors, sizes):
  """Checks that the inputs are valid for graph convolution ops.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    data: A `float` tensor with shape `[A1, ..., An, V1, V2]`.
    neighbors: A SparseTensor with the same type as `data` and with shape `[A1,
      ..., An, V1, V1]`.
    sizes: An `int` tensor of shape `[A1, ..., An]`. Optional, can be `None`.

  Raises:
    TypeError: if the input types are invalid.
    ValueError: if the input dimensions are invalid.
  """
  if not data.dtype.is_floating:
    raise TypeError("'data' must have a float type.")
  if neighbors.dtype != data.dtype:
    raise TypeError("'neighbors' and 'data' must have the same type.")
  if sizes is not None and not sizes.dtype.is_integer:
    raise TypeError("'sizes' must have an integer type.")
  if not isinstance(neighbors, tf.sparse.SparseTensor):
    raise ValueError("'neighbors' must be a SparseTensor.")

  data_ndims = data.shape.ndims
  shape.check_static(tensor=data, tensor_name="data", has_rank_greater_than=1)
  shape.check_static(
      tensor=neighbors, tensor_name="neighbors", has_rank=data_ndims)
  if not _is_dynamic_shape(tensors=(data, neighbors)):
    shape.compare_dimensions(
        tensors=(data, neighbors, neighbors),
        tensor_names=("data", "neighbors", "neighbors"),
        axes=(-2, -2, -1))
  if sizes is None:
    shape.compare_batch_dimensions(
        tensors=(data, neighbors),
        tensor_names=("data", "neighbors"),
        last_axes=-3,
        broadcast_compatible=False)
  else:
    shape.check_static(
        tensor=sizes, tensor_name="sizes", has_rank=data_ndims - 2)
    shape.compare_batch_dimensions(
        tensors=(data, neighbors, sizes),
        tensor_names=("data", "neighbors", "sizes"),
        last_axes=(-3, -3, -1),
        broadcast_compatible=False)


def check_valid_graph_pooling_input(data, pool_map, sizes):
  """Checks that the inputs are valid for graph pooling.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    data: A `float` tensor with shape `[A1, ..., An, V1, C]`.
    pool_map: A SparseTensor with the same type as `data` and with shape `[A1,
      ..., An, V2, V1]`.
    sizes: An `int` tensor of shape `[A1, ..., An, 2]`. Can be `None`.

  Raises:
    TypeError: if the input types are invalid.
    ValueError: if the input dimensions are invalid.
  """
  if not data.dtype.is_floating:
    raise TypeError("'data' must have a float type.")
  if pool_map.dtype != data.dtype:
    raise TypeError("'pool_map' and 'data' must have the same type.")
  if sizes is not None and not sizes.dtype.is_integer:
    raise TypeError("'sizes' must have an integer type.")
  if not isinstance(pool_map, tf.sparse.SparseTensor):
    raise ValueError("'pool_map' must be a SparseTensor.")

  data_ndims = data.shape.ndims
  shape.check_static(tensor=data, tensor_name="data", has_rank_greater_than=1)
  shape.check_static(
      tensor=pool_map, tensor_name="pool_map", has_rank=data_ndims)
  if not _is_dynamic_shape(tensors=(data, pool_map)):
    shape.compare_dimensions(
        tensors=(data, pool_map),
        tensor_names=("data", "pool_map"),
        axes=(-2, -1))
  if sizes is None:
    shape.compare_batch_dimensions(
        tensors=(data, pool_map),
        tensor_names=("data", "pool_map"),
        last_axes=-3,
        broadcast_compatible=False)
  else:
    shape.check_static(
        tensor=sizes, tensor_name="sizes", has_rank=data_ndims - 1)
    shape.compare_batch_dimensions(
        tensors=(data, pool_map, sizes),
        tensor_names=("data", "pool_map", "sizes"),
        last_axes=(-3, -3, -2),
        broadcast_compatible=False)


def check_valid_graph_unpooling_input(data, pool_map, sizes):
  """Checks that the inputs are valid for graph unpooling.

  Note:
    In the following, A1 to A3 are optional batch dimensions.

  Args:
    data: A `float` tensor with shape `[A1, ..., A3, V1, C]`.
    pool_map: A `SparseTensor` with the same type as `data` and with shape `[A1,
      ..., A3, V1, V2]`.
    sizes: An `int` tensor of shape `[A1, ..., A3, 2]`. Can be `None`.

  Raises:
    TypeError: if the input types are invalid.
    ValueError: if the input dimensions are invalid.
  """
  if not data.dtype.is_floating:
    raise TypeError("'data' must have a float type.")
  if pool_map.dtype != data.dtype:
    raise TypeError("'pool_map' and 'data' must have the same type.")
  if sizes is not None and not sizes.dtype.is_integer:
    raise TypeError("'sizes' must have an integer type.")
  if not isinstance(pool_map, tf.sparse.SparseTensor):
    raise ValueError("'pool_map' must be a SparseTensor.")

  data_ndims = data.shape.ndims
  shape.check_static(tensor=data, tensor_name="data", has_rank_greater_than=1)
  shape.check_static(tensor=data, tensor_name="data", has_rank_less_than=6)
  shape.check_static(
      tensor=pool_map, tensor_name="pool_map", has_rank=data_ndims)
  if not _is_dynamic_shape(tensors=(data, pool_map)):
    shape.compare_dimensions(
        tensors=(data, pool_map),
        tensor_names=("data", "pool_map"),
        axes=(-2, -2))
  if sizes is None:
    shape.compare_batch_dimensions(
        tensors=(data, pool_map),
        tensor_names=("data", "pool_map"),
        last_axes=-3,
        broadcast_compatible=False)
  else:
    shape.check_static(
        tensor=sizes, tensor_name="sizes", has_rank=data_ndims - 1)
    shape.compare_batch_dimensions(
        tensors=(data, pool_map, sizes),
        tensor_names=("data", "pool_map", "sizes"),
        last_axes=(-3, -3, -2),
        broadcast_compatible=False)


def flatten_batch_to_2d(data, sizes=None, name=None):
  """Reshapes a batch of 2d Tensors by flattening across the batch dimensions.

  Note:
    In the following, A1 to An are optional batch dimensions.

  A tensor with shape `[A1, ..., An, D1, D2]` will be reshaped to one
  with shape `[A1*...*An*D1, D2]`. This function also returns an inverse
  function that returns any tensor with shape `[A1*...*An*D1, D3]` to one
  with shape `[A1, ..., An, D1, D3]`.

  Padded inputs in dimension D1 are allowed. `sizes` determines the first
  elements from D1 to select from each batch dimension.

  Examples:
    ```python
    data = [[[1., 2.], [3., 4.]],
            [[5., 6.], [7., 8.]],
            [[9., 10.], [11., 12.]]]
    sizes = None
    output = flatten_batch_to_2d(data, size)
    print(output)
    >>> [[1., 2.], [3., 4.], [5., 6.], [7., 8.], [9., 10.], [11., 12.]]

    data = [[[1., 2.], [0., 0.]],
            [[5., 6.], [7., 8.]],
            [[9., 10.], [0., 0.]]]
    sizes = [1, 2, 1]
    output = flatten_batch_to_2d(data, size)
    print(output)
    >>> [[1., 2.], [5., 6.], [7., 8.], [9., 10.]]
    ```

  Args:
    data: A tensor with shape `[A1, ..., An, D1, D2]`.
    sizes: An `int` tensor with shape `[A1, ..., An]`. Can be `None`. `sizes[i]
      <= D1`.
    name: A name for this op. Defaults to 'utils_flatten_batch_to_2d'.

  Returns:
    A tensor with shape `[A1*...*An*D1, D2]` if `sizes == None`, otherwise a
      tensor  with shape `[sum(sizes), D2]`.
    A function that reshapes a tensor with shape `[A1*...*An*D1, D3]` to a
      tensor with shape `[A1, ..., An, D1, D3]` if `sizes == None`, otherwise
      it reshapes a tensor with shape `[sum(sizes), D3]` to one with shape
      `[A1, ..., An, ..., D1, D3]`.

  Raises:
    ValueError: if the input tensor dimensions are invalid.
  """
  with tf.compat.v1.name_scope(name, "utils_flatten_batch_to_2d",
                               [data, sizes]):
    data = tf.convert_to_tensor(value=data)
    if sizes is not None:
      sizes = tf.convert_to_tensor(value=sizes)

    if sizes is not None and not sizes.dtype.is_integer:
      raise TypeError("'sizes' must have an integer type.")
    shape.check_static(tensor=data, tensor_name="data", has_rank_greater_than=2)
    if sizes is not None:
      shape.check_static(
          tensor=sizes, tensor_name="sizes", has_rank=data.shape.ndims - 2)
      shape.compare_batch_dimensions(
          tensors=(data, sizes),
          tensor_names=("data", "sizes"),
          last_axes=(-3, -1),
          broadcast_compatible=False)

    data_shape = tf.shape(input=data)
    if sizes is None:
      flat = tf.reshape(data, shape=(-1, data_shape[-1]))

      def unflatten(flat, name=None):
        """Invert flatten_batch_to_2d."""
        with tf.compat.v1.name_scope(name, "utils_unflatten", [flat]):
          flat = tf.convert_to_tensor(value=flat)
          output_shape = tf.concat((data_shape[:-1], tf.shape(input=flat)[-1:]),
                                   axis=0)
          return tf.reshape(flat, output_shape)
    else:
      # Create a mask for the desired rows in `data` to select for flattening:
      # `mask` has shape `[A1, ..., An, D1]` and
      # `mask[a1, ..., an, :] = [True, ..., True, False, ..., False]` where
      # the number of True elements is `sizes[a1, ..., an]`.
      mask = tf.sequence_mask(sizes, data_shape[-2])
      mask_indices = tf.cast(tf.compat.v1.where(mask), tf.int32)
      flat = tf.gather_nd(params=data, indices=mask_indices)

      def unflatten(flat, name=None):
        """Invert flatten_batch_to_2d."""
        with tf.compat.v1.name_scope(name, "utils_unflatten", [flat]):
          flat = tf.convert_to_tensor(value=flat)
          output_shape = tf.concat((data_shape[:-1], tf.shape(input=flat)[-1:]),
                                   axis=0)
          return tf.scatter_nd(
              indices=mask_indices, updates=flat, shape=output_shape)

    return flat, unflatten


def unflatten_2d_to_batch(data, sizes, max_rows=None, name=None):
  r"""Reshapes a 2d Tensor into a batch of 2d Tensors.

  The `data` tensor with shape `[D1, D2]` will be mapped to a tensor with shape
  `[A1, ..., An, max_rows, D2]` where `max_rows` defaults to `max(sizes)`.
  `sizes` determines the segment of rows in the input that get mapped to a
  particular batch dimension (`sum(sizes) == D1`).

  Examples:

    ```python
    data = [[1., 2.],
            [3., 4.],
            [5., 6.],
            [7., 8.],
            [9., 10.],
            [11., 12.]]
    sizes = [2, 3, 1]

    output = unflatten_2d_to_batch(data, sizes, max_rows=None)
    print(output.shape)
    >>> [3, 3, 2]
    print(output)
    >>> [[[1., 2.],
          [3., 4.],
          [0., 0.]],
         [[5., 6.],
          [7., 8.],
          [9., 10.]],
         [[11., 12.],
          [0., 0.],
          [0., 0.]]]

    output = unflatten_2d_to_batch(data, sizes, max_rows=4)
    print(output.shape)
    >>> [3, 4, 2]
    print(output)
    >>> [[[1., 2.],
          [3., 4.],
          [0., 0.],
          [0., 0.]],
         [[5., 6.],
          [7., 8.],
          [9., 10.],
          [0., 0.]],
         [[11., 12.],
          [0., 0.],
          [0., 0.],
          [0., 0.]]]
    ```

  Args:
    data: A tensor with shape `[D1, D2]`.
    sizes: An `int` tensor with shape `[A1, ..., An]`.
    max_rows: An `int` specifying the maximum number of rows in the unflattened
      output. `max_rows >= max(sizes)`.
    name: A name for this op. Defaults to 'utils_unflatten_2d_to_batch'.

  Returns:
    A tensor with shape `[A1, A2, ..., max_rows, D2]`.
  """
  with tf.compat.v1.name_scope(name, "utils_unflatten_2d_to_batch",
                               [data, sizes]):
    data = tf.convert_to_tensor(value=data)
    sizes = tf.convert_to_tensor(value=sizes)
    if max_rows is None:
      max_rows = tf.reduce_max(input_tensor=sizes)
    else:
      max_rows = tf.convert_to_tensor(value=max_rows)

    shape.check_static(tensor=data, tensor_name="data", has_rank=2)
    if not sizes.dtype.is_integer:
      raise TypeError("'sizes' must have an integer type.")

    mask = tf.sequence_mask(sizes, max_rows)
    mask_indices = tf.cast(tf.compat.v1.where(mask), tf.int32)
    output_shape = tf.concat(
        (tf.shape(input=sizes), (max_rows,), tf.shape(input=data)[-1:]), axis=0)
    return tf.scatter_nd(indices=mask_indices, updates=data, shape=output_shape)


def convert_to_block_diag_2d(data,
                             sizes=None,
                             validate_indices=False,
                             name=None):
  """Convert a batch of 2d SparseTensors to a 2d block diagonal SparseTensor.

  Note:
    In the following, A1 to An are optional batch dimensions.

  A `SparseTensor` with dense shape `[A1, ..., An, D1, D2]` will be reshaped
  to one with shape `[A1*...*An*D1, A1*...*An*D2]`.

  Padded inputs in dims D1 and D2 are allowed. `sizes` indicates the un-padded
  shape for each inner `[D1, D2]` matrix. The additional (padded) rows and
  columns will be omitted in the block diagonal output.

  If padded (`sizes != None`), the input should not contain any sparse indices
  outside the bounds indicated by `sizes`. Setting `validate_indices=True` will
  explicitly filter any invalid sparse indices before block diagonalization.

  Args:
    data: A `SparseTensor` with dense shape `[A1, ..., An, D1, D2]`.
    sizes: A tensor with shape `[A1, ..., An, 2]`. Can be `None` (indicates no
      padding). If not `None`, `sizes` indicates the true sizes (before padding)
      of the inner dimensions of `data`.
    validate_indices: A boolean. Ignored if `sizes==None`. If True,
      out-of-bounds indices in `data` are explicitly ignored, otherwise
      out-of-bounds indices will cause undefined behavior.
    name: A name for this op. Defaults to 'utils_convert_to_block_diag_2d'.

  Returns:
    A 2d block-diagonal SparseTensor.

  Raises:
    TypeError: if the input types are invalid.
    ValueError: if the input dimensions are invalid.
  """
  with tf.compat.v1.name_scope(name, "utils_convert_to_block_diag_2d",
                               [data, sizes, validate_indices]):
    data = tf.compat.v1.convert_to_tensor_or_sparse_tensor(value=data)
    if sizes is not None:
      sizes = tf.convert_to_tensor(value=sizes)

    if not isinstance(data, tf.SparseTensor):
      raise TypeError("'data' must be a 'SparseTensor'.")
    if sizes is not None and not sizes.dtype.is_integer:
      raise TypeError("'sizes' must have an integer type.")
    shape.check_static(tensor=data, tensor_name="data", has_rank_greater_than=2)
    if sizes is not None:
      shape.check_static(
          tensor=sizes,
          tensor_name="sizes",
          has_rank=data.shape.ndims - 1,
          has_dim_equals=(-1, 2))
      shape.compare_batch_dimensions(
          tensors=(data, sizes),
          tensor_names=("data", "sizes"),
          last_axes=(-3, -2),
          broadcast_compatible=False)

    data_shape = tf.shape(input=data)
    data = tf.sparse.reshape(data, [-1, data_shape[-2], data_shape[-1]])
    indices = data.indices
    if sizes is not None:
      sizes = tf.cast(tf.reshape(sizes, shape=(-1, 2)), tf.int64)
      if validate_indices:
        in_bounds = ~tf.reduce_any(
            input_tensor=indices[:, 1:] >= tf.gather(sizes, indices[:, 0]),
            axis=-1)
        indices = tf.boolean_mask(tensor=indices, mask=in_bounds)
        values = tf.boolean_mask(tensor=data.values, mask=in_bounds)
      else:
        values = data.values
      cumsum = tf.cumsum(sizes, axis=0, exclusive=True)
      index_shift = tf.gather(cumsum, indices[:, 0])
      indices = indices[:, 1:] + index_shift
      block_diag = tf.SparseTensor(indices, values,
                                   tf.reduce_sum(input_tensor=sizes, axis=0))
    else:
      data_shape = tf.shape(input=data, out_type=tf.int64)
      index_shift = tf.expand_dims(indices[:, 0], -1) * data_shape[1:]
      indices = indices[:, 1:] + index_shift
      block_diag = tf.SparseTensor(indices, data.values,
                                   data_shape[0] * data_shape[1:])
    return block_diag


# API contains all public functions and classes.
__all__ = []



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for srgb."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import flagsaver
from absl.testing import parameterized
import numpy as np

from tensorflow_graphics.image.color_space import linear_rgb
from tensorflow_graphics.image.color_space import srgb
from tensorflow_graphics.util import test_case


class SrgbTest(test_case.TestCase):

  def test_cycle_linear_rgb_srgb_linear_rgb_for_random_input(self):
    """Tests loop from linear RGB to sRGB and back for random inputs."""
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    linear_input = np.random.uniform(size=tensor_shape + [3])

    srgb_output = srgb.from_linear_rgb(linear_input)
    linear_reverse = linear_rgb.from_srgb(srgb_output)

    self.assertAllClose(linear_input, linear_reverse)

  @parameterized.parameters(
      (((0., 0.5, 1.), (0.00312, 0.0031308, 0.00314)),
       ((0., 0.735357, 1.), (0.04031, 0.04045, 0.040567))),)
  def test_from_linear_rgb_preset(self, test_inputs, test_outputs):
    """Tests conversion from linear to sRGB color space for preset inputs."""
    self.assert_output_is_correct(srgb.from_linear_rgb, (test_inputs,),
                                  (test_outputs,))

  def test_from_linear_rgb_jacobian_random(self):
    """Tests the Jacobian of the from_linear_rgb function for random inputs."""
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    linear_random_init = np.random.uniform(size=tensor_shape + [3])

    self.assert_jacobian_is_correct_fn(srgb.from_linear_rgb,
                                       [linear_random_init])

  @parameterized.parameters((np.array((0., 0.001, 0.002)),), (np.array(
      (0.004, 0.005, 1.)),), (np.array((0.00312, 0.004, 0.00314)),))
  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_from_linear_rgb_jacobian_preset(self, inputs_init):
    """Tests the Jacobian of the from_linear_rgb function for preset inputs."""
    self.assert_jacobian_is_correct_fn(srgb.from_linear_rgb, [inputs_init])

  @parameterized.parameters(
      ((3,),),
      ((None, None, None, 3),),
  )
  def test_from_linear_rgb_exception_not_raised(self, *shape):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(srgb.from_linear_rgb, shape)

  @parameterized.parameters(
      ("must have a rank greater than 0", ()),
      ("must have exactly 3 dimensions in axis -1", (2, 3, 4)),
  )
  def test_from_linear_rgb_exception_raised(self, error_msg, *shape):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(srgb.from_linear_rgb, error_msg, shape)


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for pointnet layers."""

# pylint: disable=invalid-name

from absl.testing import parameterized
import tensorflow as tf
from tensorflow_graphics.nn.layer.pointnet import ClassificationHead
from tensorflow_graphics.nn.layer.pointnet import PointNetConv2Layer
from tensorflow_graphics.nn.layer.pointnet import PointNetDenseLayer
from tensorflow_graphics.nn.layer.pointnet import PointNetVanillaClassifier
from tensorflow_graphics.nn.layer.pointnet import VanillaEncoder
from tensorflow_graphics.util import test_case


class RandomForwardExecutionTest(test_case.TestCase):

  @parameterized.parameters(
      ((32, 2048, 1, 3), (32), (.5), True),
      ((32, 2048, 1, 3), (32), (.5), False),
      ((32, 2048, 1, 2), (16), (.99), True),
  )
  def test_conv2(self, input_shape, channels, momentum, training):
    B, N, X, _ = input_shape
    inputs = tf.random.uniform(input_shape)
    layer = PointNetConv2Layer(channels, momentum)
    outputs = layer(inputs, training=training)
    assert outputs.shape == (B, N, X, channels)

  @parameterized.parameters(
      ((32, 1024), (40), (.5), True),
      ((32, 2048), (20), (.5), False),
      ((32, 512), (10), (.99), True),
  )
  def test_dense(self, input_shape, channels, momentum, training):
    B, _ = input_shape
    inputs = tf.random.uniform(input_shape)
    layer = PointNetDenseLayer(channels, momentum)
    outputs = layer(inputs, training=training)
    assert outputs.shape == (B, channels)

  @parameterized.parameters(
      ((32, 2048, 3), (.9), True),
      ((32, 2048, 2), (.5), False),
      ((32, 2048, 3), (.99), True),
  )
  def test_vanilla_encoder(self, input_shape, momentum, training):
    B = input_shape[0]
    inputs = tf.random.uniform(input_shape)
    encoder = VanillaEncoder(momentum)
    outputs = encoder(inputs, training=training)
    assert outputs.shape == (B, 1024)

  @parameterized.parameters(
      ((16, 1024), (20), (.9), True),
      ((8, 2048), (40), (.5), False),
      ((32, 512), (10), (.99), True),
  )
  def test_classification_head(self, input_shape, num_classes, momentum,
                               training):
    B = input_shape[0]
    inputs = tf.random.uniform(input_shape)
    head = ClassificationHead(num_classes, momentum)
    outputs = head(inputs, training=training)
    assert outputs.shape == (B, num_classes)

  @parameterized.parameters(
      ((32, 1024, 3), 40, True),
      ((32, 1024, 2), 40, False),
      ((16, 2048, 3), 20, True),
      ((16, 2048, 2), 20, False),
  )
  def test_vanilla_classifier(self, input_shape, num_classes, training):
    B = input_shape[0]
    C = num_classes
    inputs = tf.random.uniform(input_shape)
    model = PointNetVanillaClassifier(num_classes, momentum=.5)
    logits = model(inputs, training)
    assert logits.shape == (B, C)
    labels = tf.random.uniform((B,), minval=0, maxval=C, dtype=tf.int64)
    PointNetVanillaClassifier.loss(labels, logits)


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""This module implements orthographic camera functionalities.

An orthographic camera represents three-dimensional objects in two dimensions
by parallel projection, in which the projection lines are parallel to the
camera axis. The camera axis is the line perpendicular to the image plane
starting at the camera center.

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from tensorflow_graphics.util import export_api
from tensorflow_graphics.util import shape


def project(point_3d, name=None):
  r"""Projects a 3d point onto the 2d camera plane.

  Projects a 3d point \\((x, y, z)\\) to a 2d point \\((x', y')\\) onto the
  image plane, with

  $$
  \begin{matrix}
  x' = x, & y' = y.
  \end{matrix}
  $$

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    point_3d: A tensor of shape `[A1, ..., An, 3]`, where the last dimension
      represents a 3d point to project.
    name: A name for this op that defaults to "orthographic_project".

  Returns:
    A tensor of shape `[A1, ..., An, 2]`, where the last dimension represents
    a 2d point.

  Raises:
    ValueError: If the shape of `point_3d` is not supported.
  """
  with tf.compat.v1.name_scope(name, "orthographic_project", [point_3d]):
    point_3d = tf.convert_to_tensor(value=point_3d)

    shape.check_static(
        tensor=point_3d, tensor_name="point_3d", has_dim_equals=(-1, 3))

    point_xy, _ = tf.compat.v1.split(point_3d, (2, 1), axis=-1)
    return point_xy


def ray(point_2d, name=None):
  r"""Computes the 3d ray for a 2d point (the z component of the ray is 1).

  Computes the 3d ray \\((r_x, r_y, 1)\\) for a 2d point \\((x', y')\\) on the
  image plane. For an orthographic camera the rays are constant over the image
  plane with

  $$
  \begin{matrix}
  r_x = 0, & r_y = 0, & z = 1.
  \end{matrix}
  $$

  Note: In the following, A1 to An are optional batch dimensions.

  Args:
    point_2d: A tensor of shape `[A1, ..., An, 2]`, where the last dimension
      represents a 2d point.
    name: A name for this op that defaults to "orthographic_ray".

  Returns:
    A tensor of shape `[A1, ..., An, 3]`, where the last dimension represents
    a 3d ray.

  Raises:
    ValueError: If the shape of `point_2d` is not supported.
  """
  with tf.compat.v1.name_scope(name, "orthographic_ray", [point_2d]):
    point_2d = tf.convert_to_tensor(value=point_2d)

    shape.check_static(
        tensor=point_2d, tensor_name="point_2d", has_dim_equals=(-1, 2))

    ones = tf.ones_like(point_2d[..., :1])
    # point_2d is multiplied by zero to ensure it has defined gradients.
    return tf.concat((point_2d * 0.0, ones), axis=-1)


def unproject(point_2d, depth, name=None):
  r"""Unprojects a 2d point in 3d.

  Unprojects a 2d point \\((x', y')\\) to a 3d point \\((x, y, z)\\) given its
  depth \\(z\\), with

  $$
  \begin{matrix}
  x = x', & y = y', & z = z.
  \end{matrix}
  $$

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    point_2d: A tensor of shape `[A1, ..., An, 2]`, where the last dimension
      represents a 2d point to unproject.
    depth: A tensor of shape `[A1, ..., An, 1]`, where the last dimension
      represents the depth of a 2d point.
    name: A name for this op that defaults to "orthographic_unproject".

  Returns:
    A tensor of shape `[A1, ..., An, 3]`, where the last dimension represents
    a 3d point.

  Raises:
    ValueError: If the shape of `point_2d`, `depth` is not supported.
  """
  with tf.compat.v1.name_scope(name, "orthographic_unproject",
                               [point_2d, depth]):
    point_2d = tf.convert_to_tensor(value=point_2d)
    depth = tf.convert_to_tensor(value=depth)

    shape.check_static(
        tensor=point_2d, tensor_name="point_2d", has_dim_equals=(-1, 2))
    shape.check_static(
        tensor=depth, tensor_name="depth", has_dim_equals=(-1, 1))
    shape.compare_batch_dimensions(
        tensors=(point_2d, depth),
        tensor_names=("point_2d", "depth"),
        last_axes=-2,
        broadcast_compatible=False)

    return tf.concat((point_2d, depth), axis=-1)


# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tensorflow utility functions to compute normals on meshes."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np

import tensorflow as tf

from tensorflow_graphics.geometry.representation import triangle
from tensorflow_graphics.util import export_api
from tensorflow_graphics.util import safe_ops
from tensorflow_graphics.util import shape


def gather_faces(vertices, indices, name=None):
  """Gather corresponding vertices for each face.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    vertices: A tensor of shape `[A1, ..., An, V, D]`, where `V` is the number
      of vertices and `D` the dimensionality of each vertex. The rank of this
      tensor should be at least 2.
    indices: A tensor of shape `[A1, ..., An, F, M]`, where `F` is the number of
      faces, and `M` is the number of vertices per face. The rank of this tensor
      should be at least 2.
    name: A name for this op. Defaults to "normals_gather_faces".

  Returns:
    A tensor of shape `[A1, ..., An, F, M, D]` containing the vertices of each
    face.

  Raises:
    ValueError: If the shape of `vertices` or `indices` is not supported.
  """
  with tf.compat.v1.name_scope(name, "normals_gather_faces",
                               [vertices, indices]):
    vertices = tf.convert_to_tensor(value=vertices)
    indices = tf.convert_to_tensor(value=indices)

    shape.check_static(
        tensor=vertices, tensor_name="vertices", has_rank_greater_than=1)
    shape.check_static(
        tensor=indices, tensor_name="indices", has_rank_greater_than=1)
    shape.compare_batch_dimensions(
        tensors=(vertices, indices),
        last_axes=(-3, -3),
        broadcast_compatible=False)

    if hasattr(tf, "batch_gather"):
      expanded_vertices = tf.expand_dims(vertices, axis=-3)
      broadcasted_shape = tf.concat([tf.shape(input=indices)[:-1],
                                     tf.shape(input=vertices)[-2:]],
                                    axis=-1)
      broadcasted_vertices = tf.broadcast_to(
          expanded_vertices,
          broadcasted_shape)
      return tf.compat.v1.batch_gather(broadcasted_vertices, indices)
    else:
      return tf.gather(
          vertices, indices, axis=-2, batch_dims=indices.shape.ndims - 2)


def face_normals(faces, clockwise=True, normalize=True, name=None):
  """Computes face normals for meshes.

  This function supports planar convex polygon faces. Note that for
  non-triangular faces, this function uses the first 3 vertices of each
  face to calculate the face normal.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    faces: A tensor of shape `[A1, ..., An, M, 3]`, which stores vertices
      positions of each face, where M is the number of vertices of each face.
      The rank of this tensor should be at least 2.
    clockwise: Winding order to determine front-facing faces. The order of
      vertices should be either clockwise or counterclockwise.
    normalize: A `bool` defining whether output normals are normalized.
    name: A name for this op. Defaults to "normals_face_normals".

  Returns:
    A tensor of shape `[A1, ..., An, 3]` containing the face normals.

  Raises:
    ValueError: If the shape of `vertices`, `faces` is not supported.
  """
  with tf.compat.v1.name_scope(name, "normals_face_normals", [faces]):
    faces = tf.convert_to_tensor(value=faces)

    shape.check_static(
        tensor=faces,
        tensor_name="faces",
        has_rank_greater_than=1,
        has_dim_equals=(-1, 3),
        has_dim_greater_than=(-2, 2))

    vertices = tf.unstack(faces, axis=-2)
    vertices = vertices[:3]
    return triangle.normal(*vertices, clockwise=clockwise, normalize=normalize)


def vertex_normals(vertices, indices, clockwise=True, name=None):
  """Computes vertex normals from a mesh.

  This function computes vertex normals as the weighted sum of the adjacent
  face normals, where the weights correspond to the area of each face. This
  function supports planar convex polygon faces. For non-triangular meshes,
  this function converts them into triangular meshes to calculate vertex
  normals.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    vertices: A tensor of shape `[A1, ..., An, V, 3]`, where V is the number of
      vertices.
    indices: A tensor of shape `[A1, ..., An, F, M]`, where F is the number of
      faces and M is the number of vertices per face.
    clockwise: Winding order to determine front-facing faces. The order of
      vertices should be either clockwise or counterclockwise.
    name: A name for this op. Defaults to "normals_vertex_normals".

  Returns:
    A tensor of shape `[A1, ..., An, V, 3]` containing vertex normals. If
    vertices and indices have different batch dimensions, this function
    broadcasts them into the same batch dimensions and the output batch
    dimensions are the broadcasted.

  Raises:
    ValueError: If the shape of `vertices`, `indices` is not supported.
  """
  with tf.compat.v1.name_scope(name, "normals_vertex_normals",
                               [vertices, indices]):
    vertices = tf.convert_to_tensor(value=vertices)
    indices = tf.convert_to_tensor(value=indices)

    shape.check_static(
        tensor=vertices,
        tensor_name="vertices",
        has_rank_greater_than=1,
        has_dim_equals=(-1, 3))
    shape.check_static(
        tensor=indices,
        tensor_name="indices",
        has_rank_greater_than=1,
        has_dim_greater_than=(-1, 2))
    shape.compare_batch_dimensions(
        tensors=(vertices, indices),
        last_axes=(-3, -3),
        broadcast_compatible=True)

    shape_indices = indices.shape.as_list()
    if None in shape_indices[:-2]:
      raise ValueError("'indices' must have specified batch dimensions.")
    common_batch_dims = shape.get_broadcasted_shape(vertices.shape[:-2],
                                                    indices.shape[:-2])
    vertices_repeat = [
        common_batch_dims[x] // vertices.shape.as_list()[x]
        for x in range(len(common_batch_dims))
    ]
    indices_repeat = [
        common_batch_dims[x] // shape_indices[x]
        for x in range(len(common_batch_dims))
    ]
    vertices = tf.tile(
        vertices, vertices_repeat + [1, 1], name="vertices_broadcast")
    indices = tf.tile(
        indices, indices_repeat + [1, 1], name="indices_broadcast")

    # Triangulate non-triangular faces.
    if shape_indices[-1] > 3:
      triangle_indices = []
      for i in range(1, shape_indices[-1] - 1):
        triangle_indices.append(
            tf.concat((indices[..., 0:1], indices[..., i:i + 2]), axis=-1))
      indices = tf.concat(triangle_indices, axis=-2)
      shape_indices = indices.shape.as_list()

    face_vertices = gather_faces(vertices, indices)
    # Use unnormalized face normals to scale normals by area.
    mesh_face_normals = face_normals(
        face_vertices, clockwise=clockwise, normalize=False)

    if vertices.shape.ndims > 2:
      outer_indices = np.meshgrid(
          *[np.arange(i) for i in shape_indices[:-2]],
          sparse=False,
          indexing="ij")
      outer_indices = [np.expand_dims(i, axis=-1) for i in outer_indices]
      outer_indices = np.concatenate(outer_indices, axis=-1)
      outer_indices = np.expand_dims(outer_indices, axis=-2)
      outer_indices = tf.constant(outer_indices, dtype=tf.int32)
      outer_indices = tf.tile(outer_indices, [1] * len(shape_indices[:-2]) +
                              [tf.shape(input=indices)[-2]] + [1])
      unnormalized_vertex_normals = tf.zeros_like(vertices)
      for i in range(shape_indices[-1]):
        scatter_indices = tf.concat([outer_indices, indices[..., i:i + 1]],
                                    axis=-1)
        unnormalized_vertex_normals = tf.compat.v1.tensor_scatter_add(
            unnormalized_vertex_normals, scatter_indices, mesh_face_normals)
    else:
      unnormalized_vertex_normals = tf.zeros_like(vertices)
      for i in range(shape_indices[-1]):
        unnormalized_vertex_normals = tf.compat.v1.tensor_scatter_add(
            unnormalized_vertex_normals, indices[..., i:i + 1],
            mesh_face_normals)

    vector_norms = tf.sqrt(
        tf.reduce_sum(
            input_tensor=unnormalized_vertex_normals**2, axis=-1,
            keepdims=True))
    return safe_ops.safe_unsigned_div(unnormalized_vertex_normals, vector_norms)


# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for lambertian reflectance."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math
import sys

from absl.testing import flagsaver
from absl.testing import parameterized
import numpy as np
import tensorflow as tf

from tensorflow_graphics.rendering.reflectance import lambertian
from tensorflow_graphics.util import test_case


class LambertianTest(test_case.TestCase):

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_brdf_jacobian_random(self):
    """Tests the Jacobian of brdf."""
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    direction_incoming_light_init = np.random.uniform(
        1.0, 1.0, size=tensor_shape + [3])
    direction_outgoing_light_init = np.random.uniform(
        1.0, 1.0, size=tensor_shape + [3])
    surface_normal_init = np.random.uniform(1.0, 1.0, size=tensor_shape + [3])
    albedo_init = np.random.random(tensor_shape + [3])

    def lamertian_brdf_fn(albedo):
      return lambertian.brdf(direction_incoming_light_init,
                             direction_outgoing_light_init, surface_normal_init,
                             albedo)

    self.assert_jacobian_is_correct_fn(lamertian_brdf_fn, [albedo_init])

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_brdf_jacobian_preset(self):
    direction_incoming_light_init = np.array((0.0, 1.0, 0.0))
    direction_outgoing_light_init = np.array((0.0, 1.0, 0.0))
    surface_normal_init = np.array((1.0, 0.0, 0.0))
    albedo_init = np.array((1.0, 1.0, 1.0))

    def lamertian_brdf_fn(albedo):
      return lambertian.brdf(direction_incoming_light_init,
                             direction_outgoing_light_init, surface_normal_init,
                             albedo)

    self.assert_jacobian_is_correct_fn(lamertian_brdf_fn, [albedo_init])

  @parameterized.parameters(
      (-1.0, 1.0, 1.0 / math.pi),
      (1.0, 1.0, 0.0),
      (-1.0, -1.0, 0.0),
      (1.0, -1.0, 0.0),
  )
  def test_brdf_random(self, incoming_yz, outgoing_yz, ratio):
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    albedo = np.random.uniform(low=0.0, high=1.0, size=tensor_shape + [3])
    direction_incoming_light = np.random.uniform(
        low=-1.0, high=1.0, size=tensor_shape + [3])
    direction_outgoing_light = np.random.uniform(
        low=-1.0, high=1.0, size=tensor_shape + [3])
    surface_normal = np.array((0.0, 1.0, 1.0))
    direction_incoming_light[..., 1:3] = incoming_yz
    direction_outgoing_light[..., 1:3] = outgoing_yz
    direction_incoming_light = direction_incoming_light / np.linalg.norm(
        direction_incoming_light, axis=-1, keepdims=True)
    direction_outgoing_light = direction_outgoing_light / np.linalg.norm(
        direction_outgoing_light, axis=-1, keepdims=True)
    surface_normal = surface_normal / np.linalg.norm(
        surface_normal, axis=-1, keepdims=True)

    gt = albedo * ratio
    pred = lambertian.brdf(direction_incoming_light, direction_outgoing_light,
                           surface_normal, albedo)

    self.assertAllClose(gt, pred)

  def test_brdf_exceptions_raised(self):
    """Tests that the exceptions are raised correctly."""
    direction_incoming_light = np.random.uniform(-1.0, 1.0, size=(3,))
    direction_outgoing_light = np.random.uniform(-1.0, 1.0, size=(3,))
    surface_normal = np.random.uniform(-1.0, 1.0, size=(3,))
    albedo = np.random.uniform(0.0, 1.0, (3,))

    with self.subTest(name="assert_on_direction_incoming_light_not_normalized"):
      with self.assertRaises(tf.errors.InvalidArgumentError):
        self.evaluate(
            lambertian.brdf(direction_incoming_light, direction_outgoing_light,
                            surface_normal, albedo))

    direction_incoming_light /= np.linalg.norm(
        direction_incoming_light, axis=-1)
    with self.subTest(name="assert_on_direction_outgoing_light_not_normalized"):
      with self.assertRaises(tf.errors.InvalidArgumentError):
        self.evaluate(
            lambertian.brdf(direction_incoming_light, direction_outgoing_light,
                            surface_normal, albedo))

    direction_outgoing_light /= np.linalg.norm(
        direction_outgoing_light, axis=-1)
    with self.subTest(name="assert_on_surface_normal_not_normalized"):
      with self.assertRaises(tf.errors.InvalidArgumentError):
        self.evaluate(
            lambertian.brdf(direction_incoming_light, direction_outgoing_light,
                            surface_normal, albedo))

    surface_normal /= np.linalg.norm(surface_normal, axis=-1)
    with self.subTest(name="assert_on_albedo_not_normalized"):
      albedo = np.random.uniform(-10.0, -sys.float_info.epsilon, (3,))

      with self.assertRaises(tf.errors.InvalidArgumentError):
        self.evaluate(
            lambertian.brdf(direction_incoming_light, direction_outgoing_light,
                            surface_normal, albedo))

      albedo = np.random.uniform(sys.float_info.epsilon, 10.0, (3,))
      with self.assertRaises(tf.errors.InvalidArgumentError):
        self.evaluate(
            lambertian.brdf(direction_incoming_light, direction_outgoing_light,
                            surface_normal, albedo))

  @parameterized.parameters(
      ((3,), (3,), (3,), (3,)),
      ((None, 3), (None, 3), (None, 3), (None, 3)),
      ((1, 3), (1, 3), (1, 3), (1, 3)),
      ((2, 3), (2, 3), (2, 3), (2, 3)),
      ((3,), (1, 3), (1, 2, 3), (1, 3)),
      ((3,), (1, 3), (1, 2, 3), (1, 2, 2, 3)),
      ((1, 2, 2, 3), (1, 2, 3), (1, 3), (3,)),
  )
  def test_brdf_shape_exception_not_raised(self, *shape):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(lambertian.brdf, shape)

  @parameterized.parameters(
      ("must have exactly 3 dimensions in axis -1", (1,), (3,), (3,), (3,)),
      ("must have exactly 3 dimensions in axis -1", (2,), (3,), (3,), (3,)),
      ("must have exactly 3 dimensions in axis -1", (4,), (3,), (3,), (3,)),
      ("must have exactly 3 dimensions in axis -1", (3,), (1,), (3,), (3,)),
      ("must have exactly 3 dimensions in axis -1", (3,), (2,), (3,), (3,)),
      ("must have exactly 3 dimensions in axis -1", (3,), (4,), (3,), (3,)),
      ("must have exactly 3 dimensions in axis -1", (3,), (3,), (1,), (3,)),
      ("must have exactly 3 dimensions in axis -1", (3,), (3,), (2,), (3,)),
      ("must have exactly 3 dimensions in axis -1", (3,), (3,), (4,), (3,)),
      ("must have exactly 3 dimensions in axis -1", (3,), (3,), (3,), (4,)),
      ("must have exactly 3 dimensions in axis -1", (3,), (3,), (3,), (2,)),
      ("must have exactly 3 dimensions in axis -1", (3,), (3,), (3,), (1,)),
      ("Not all batch dimensions are broadcast-compatible.", (2, 3), (3, 3),
       (3,), (3,)),
  )
  def test_brdf_shape_exception_raised(self, error_msg, *shape):
    """Tests that the shape exception is raised."""
    self.assert_exception_is_raised(lambertian.brdf, error_msg, shape)


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Setup for pip package.

Usage:
  python setup.py sdist bdist_wheel [--nightly]

Note:
  assumes setup.py file lives at the root of the project.
"""

import datetime
import os
import sys
import setuptools

# --- Name to use in PyPi
if "--nightly" in sys.argv:
  sys.argv.remove("--nightly")
  NAME = "tfg-nightly"
else:
  NAME = "tensorflow-graphics"

# --- compute the version (for both nightly and normal)
now = datetime.datetime.now()
VERSION = "{}.{}.{}".format(now.year, now.month, now.day)
curr_path = os.path.dirname(__file__)
ini_file_path = os.path.join(curr_path, "tensorflow_graphics/__init__.py")
ini_file_lines = list(open(ini_file_path))
with open(ini_file_path, "w") as f:
  for line in ini_file_lines:
    f.write(line.replace("__version__ = \"HEAD\"",
                         "__version__ = \"{}\"".format(VERSION)))

# --- Extract the dependencies
REQS = [line.strip() for line in open("requirements.txt")]
INSTALL_PACKAGES = [line for line in REQS if not line.startswith("#")]

# --- Build the whl file
setuptools.setup(
    name=NAME,
    version=VERSION,
    description=("A library that contains well defined, reusable and cleanly "
                 "written graphics related ops and utility functions for "
                 "TensorFlow."),
    long_description="",
    url="https://github.com/tensorflow/graphics",
    author="Google LLC",
    author_email="tf-graphics-eng@google.com",
    install_requires=INSTALL_PACKAGES,
    include_package_data=True,
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    license="Apache 2.0",
    keywords=[
        "machine learning",
        "tensorflow",
        "graphics",
        "geometry",
        "3D",
    ],
)



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for safe_ops."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import parameterized
import numpy as np
import tensorflow as tf

from tensorflow_graphics.util import asserts
from tensorflow_graphics.util import safe_ops
from tensorflow_graphics.util import test_case


def _pick_random_vector():
  """Creates a random vector with random shape."""
  tensor_size = np.random.randint(3)
  tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
  return np.random.normal(size=tensor_shape + [4])


class SafeOpsTest(test_case.TestCase):

  @parameterized.parameters(tf.float16, tf.float32, tf.float64)
  def test_safe_unsigned_div_exception_not_raised(self, dtype):
    """Checks that unsigned division does not cause Inf values."""
    vector = tf.convert_to_tensor(value=_pick_random_vector(), dtype=dtype)
    zero_vector = tf.zeros_like(vector)

    self.assert_exception_is_not_raised(
        safe_ops.safe_unsigned_div,
        shapes=[],
        a=tf.norm(tensor=vector),
        b=tf.norm(tensor=zero_vector))

  @parameterized.parameters(tf.float16, tf.float32, tf.float64)
  def test_safe_unsigned_div_exception_raised(self, dtype):
    """Checks that unsigned division causes Inf values for zero eps."""
    vector = tf.convert_to_tensor(value=_pick_random_vector(), dtype=dtype)
    zero_vector = tf.zeros_like(vector)

    with self.assertRaises(tf.errors.InvalidArgumentError):
      self.evaluate(
          safe_ops.safe_unsigned_div(
              tf.norm(tensor=vector), tf.norm(tensor=zero_vector), eps=0.0))

  @parameterized.parameters(tf.float16, tf.float32, tf.float64)
  def test_safe_signed_div_exception_not_raised(self, dtype):
    """Checks that signed division does not cause Inf values."""
    vector = tf.convert_to_tensor(value=_pick_random_vector(), dtype=dtype)
    zero_vector = tf.zeros_like(vector)

    self.assert_exception_is_not_raised(
        safe_ops.safe_signed_div,
        shapes=[],
        a=tf.norm(tensor=vector),
        b=tf.sin(zero_vector))

  @parameterized.parameters(tf.float16, tf.float32, tf.float64)
  def test_safe_signed_div_exception_raised(self, dtype):
    """Checks that signed division causes Inf values for zero eps."""
    vector = tf.convert_to_tensor(value=_pick_random_vector(), dtype=dtype)
    zero_vector = tf.zeros_like(vector)

    with self.assertRaises(tf.errors.InvalidArgumentError):
      self.evaluate(
          safe_ops.safe_unsigned_div(
              tf.norm(tensor=vector), tf.sin(zero_vector), eps=0.0))

  @parameterized.parameters(tf.float32, tf.float64)
  def test_safe_shrink_exception_not_raised(self, dtype):
    """Checks whether safe shrinking makes tensor safe for tf.acos(x)."""
    tensor = tf.convert_to_tensor(value=_pick_random_vector(), dtype=dtype)
    tensor = tensor * tensor
    norm_tensor = tensor / tf.reduce_max(
        input_tensor=tensor, axis=-1, keepdims=True)
    eps = asserts.select_eps_for_addition(dtype)
    norm_tensor += eps

    safe_tensor = safe_ops.safe_shrink(norm_tensor, -1.0, 1.0)
    self.assert_exception_is_not_raised(tf.acos, shapes=[], x=safe_tensor)

  @parameterized.parameters(tf.float32, tf.float64)
  def test_safe_shrink_exception_raised(self, dtype):
    """Checks whether safe shrinking fails when eps is zero."""
    tensor = tf.convert_to_tensor(value=_pick_random_vector(), dtype=dtype)
    tensor = tensor * tensor
    norm_tensor = tensor / tf.reduce_max(
        input_tensor=tensor, axis=-1, keepdims=True)
    eps = asserts.select_eps_for_addition(dtype)
    norm_tensor += eps

    with self.assertRaises(tf.errors.InvalidArgumentError):
      self.evaluate(safe_ops.safe_shrink(norm_tensor, -1.0, 1.0, eps=0.0))

  def test_safe_sinpx_div_sinx(self):
    """Tests for edge cases and continuity for sin(px)/sin(x)."""
    angle_step = np.pi / 16.0

    with self.subTest(name="all_angles"):
      theta = tf.range(-2.0 * np.pi, 2.0 * np.pi + angle_step / 2.0, angle_step)
      factor = np.random.uniform(size=(1,))

      division = safe_ops.safe_sinpx_div_sinx(theta, factor)
      division_l = safe_ops.safe_sinpx_div_sinx(theta + 1e-10, factor)
      division_r = safe_ops.safe_sinpx_div_sinx(theta - 1e-10, factor)

      self.assertAllClose(division, division_l, rtol=1e-9)
      self.assertAllClose(division, division_r, rtol=1e-9)

    with self.subTest(name="theta_is_zero"):
      theta = 0.0
      factor = tf.range(0.0, 1.0, 0.001)

      division = safe_ops.safe_sinpx_div_sinx(theta, factor)
      division_l = safe_ops.safe_sinpx_div_sinx(theta + 1e-10, factor)
      division_r = safe_ops.safe_sinpx_div_sinx(theta - 1e-10, factor)

      self.assertAllClose(division, division_l, atol=1e-9)
      self.assertAllClose(division, division_r, atol=1e-9)
      # According to l'Hopital rule, limit should be factor
      self.assertAllClose(division, factor, atol=1e-9)

    with self.subTest(name="theta_is_pi"):
      theta = np.pi
      factor = tf.range(0.0, 1.001, 0.001)

      division = safe_ops.safe_sinpx_div_sinx(theta, factor)
      division_l = safe_ops.safe_sinpx_div_sinx(theta + 1e-10, factor)
      division_r = safe_ops.safe_sinpx_div_sinx(theta - 1e-10, factor)

      self.assertAllClose(division, division_l, atol=1e-9)
      self.assertAllClose(division, division_r, atol=1e-9)

  def test_safe_cospx_div_cosx(self):
    """Tests for edge cases and continuity for cos(px)/cos(x)."""
    angle_step = np.pi / 16.0

    with self.subTest(name="all_angles"):
      theta = tf.range(-2.0 * np.pi, 2.0 * np.pi + angle_step / 2.0, angle_step)
      factor = np.random.uniform(size=(1,))

      division = safe_ops.safe_cospx_div_cosx(theta, factor)
      division_l = safe_ops.safe_cospx_div_cosx(theta + 1e-10, factor)
      division_r = safe_ops.safe_cospx_div_cosx(theta - 1e-10, factor)

      self.assertAllClose(division, division_l, rtol=1e-9)
      self.assertAllClose(division, division_r, rtol=1e-9)

    with self.subTest(name="theta_is_pi_over_two"):
      theta = np.pi / 2.0
      factor = tf.constant(1.0)

      division = safe_ops.safe_cospx_div_cosx(theta, factor)
      division_l = safe_ops.safe_cospx_div_cosx(theta + 1e-10, factor)
      division_r = safe_ops.safe_cospx_div_cosx(theta - 1e-10, factor)

      self.assertAllClose(division, division_l, atol=1e-9)
      self.assertAllClose(division, division_r, atol=1e-9)
      # According to l'Hopital rule, limit should be 1.0
      self.assertAllClose(division, 1.0, atol=1e-9)

    with self.subTest(name="theta_is_three_pi_over_two"):
      theta = np.pi * 3.0 / 2.0
      factor = tf.range(0.0, 1.001, 0.001)

      division = safe_ops.safe_cospx_div_cosx(theta, factor)
      division_l = safe_ops.safe_cospx_div_cosx(theta + 1e-10, factor)
      division_r = safe_ops.safe_cospx_div_cosx(theta - 1e-10, factor)

      self.assertAllClose(division, division_l, atol=1e-9)
      self.assertAllClose(division, division_r, atol=1e-9)


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module implements utility functions for meshes."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np


def extract_unique_edges_from_triangular_mesh(faces, directed_edges=False):
  """Extracts all the unique edges using the faces of a mesh.

  Args:
    faces: A numpy.ndarray of shape [T, 3], where T is the number of triangular
      faces in the mesh. Each entry in this array describes the index of a
      vertex in the mesh.
    directed_edges: A boolean flag, whether to treat an edge as directed or
      undirected.  If (i, j) is an edge in the mesh and directed_edges is True,
      then both (i, j) and (j, i) are returned in the list of edges. If (i, j)
      is an edge in the mesh and directed_edges is False, then one of (i, j) or
      (j, i) is returned.

  Returns:
    A numpy.ndarray of shape [E, 2], where E is the number of edges in
    the mesh.


    For eg: given faces = [[0, 1, 2], [0, 1, 3]], then
      for directed_edges = False, one valid output is
        [[0, 1], [0, 2], [0, 3], [1, 2], [3, 1]]
      for directed_edges = True, one valid output is
        [[0, 1], [0, 2], [0, 3], [1, 0], [1, 2], [1, 3],
         [2, 0], [2, 1], [3, 0], [3, 1]]


  Raises:
    ValueError: If `faces` is not a numpy.ndarray or if its shape is not
      supported.
  """
  if not isinstance(faces, np.ndarray):
    raise ValueError("'faces' must be a numpy.ndarray.")
  faces_shape = faces.shape
  faces_rank = len(faces_shape)
  if faces_rank != 2:
    raise ValueError(
        "'faces' must have a rank equal to 2, but it has rank {} and shape {}."
        .format(faces_rank, faces_shape))
  if faces_shape[1] != 3:
    raise ValueError(
        "'faces' must have exactly 3 dimensions in the last axis, but it has {}"
        " dimensions and is of shape {}.".format(faces_shape[1], faces_shape))
  edges = np.concatenate([faces[:, 0:2], faces[:, 1:3], faces[:, [2, 0]]],
                         axis=0)
  if directed_edges:
    edges = np.concatenate([edges, np.flip(edges, axis=-1)])
  unique_edges = np.unique(edges, axis=0)
  return unique_edges


def get_degree_based_edge_weights(edges, dtype=np.float32):
  r"""Computes vertex degree based weights for edges of a mesh.

  The degree (valence) of a vertex is number of edges incident on the vertex.
  The weight for an edge $w_{ij}$ connecting vertex $v_i$ and vertex $v_j$
  is defined as,
  $$
  w_{ij} = 1.0 / degree(v_i)
  \sum_{j} w_{ij} = 1
  $$

  Args:
    edges: A numpy.ndarray of shape [E, 2], where E is the number of directed
      edges in the mesh.
    dtype: A numpy float data type. The output weights are of data type dtype.

  Returns:
    weights: A dtype numpy.ndarray of shape [E,] denoting edge weights.

  Raises:
    ValueError: If `edges` is not a numpy.ndarray or if its shape is not
      supported, or dtype is not a float type.

  """
  if not isinstance(dtype(1), np.floating):
    raise ValueError("'dtype' must be a numpy float type.")
  if not isinstance(edges, np.ndarray):
    raise ValueError("'edges' must be a numpy.ndarray.")
  edges_shape = edges.shape
  edges_rank = len(edges_shape)
  if edges_rank != 2:
    raise ValueError(
        "'edges' must have a rank equal to 2, but it has rank {} and shape {}."
        .format(edges_rank, edges_shape))
  if edges_shape[1] != 2:
    raise ValueError(
        "'edges' must have exactly 2 dimensions in the last axis, but it has {}"
        " dimensions and is of shape {}.".format(edges_shape[1], edges_shape))
  degree = np.bincount(edges[:, 0])
  rep_degree = degree[edges[:, 0]]
  weights = 1.0 / rep_degree.astype(dtype)
  return weights


# API contains all public functions and classes.
__all__ = []



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for uniform mesh sampler."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import parameterized
import numpy as np
import tensorflow as tf

from tensorflow_graphics.geometry.representation.mesh import sampler
from tensorflow_graphics.geometry.representation.mesh.tests import mesh_test_utils
from tensorflow_graphics.util import test_case


class MeshSamplerTest(test_case.TestCase):

  def setUp(self):
    """Sets up default parameters."""
    super(MeshSamplerTest, self).setUp()
    self._test_sigma_compare_tolerance = 4.0

  def compare_poisson_equivalence(self, expected, actual):
    """Performs equivalence check on Poisson-distributed random variables."""
    delta = np.sqrt(expected) * self._test_sigma_compare_tolerance
    self.assertAllClose(expected, actual, atol=delta)

  # Tests for generate_random_face_indices
  @parameterized.parameters(
      (((), (2,)), (tf.int32, tf.float32)),
      (((), (None, 1)), (tf.int64, tf.float32)),
      (((), (None, 3, 4)), (tf.int64, tf.float64)),
  )
  def test_random_face_indices_exception_not_raised(self, shapes, dtypes):
    """Tests that shape exceptions are not raised for random_face_indices."""
    self.assert_exception_is_not_raised(sampler.generate_random_face_indices,
                                        shapes, dtypes)

  @parameterized.parameters(
      ("face_weights must have a rank greater than 0.", (), ()),
      ("num_samples must have a rank of 0.", (None,), (1, 2)),
      ("num_samples must have a rank of 0.", (4, 2), (1,)),
  )
  def test_random_face_indices_shape_exception_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised for random_face_indices."""
    self.assert_exception_is_raised(
        sampler.generate_random_face_indices, error_msg, shapes)

  def test_negative_weights_random_face_indices_exception(self):
    """Test for exception for random_face_indices with negative weights."""
    face_wts = np.array([0.1, -0.1], dtype=np.float32)
    num_samples = 10
    error_msg = "Condition x >= y did not hold."
    with self.assertRaisesRegexp(tf.errors.InvalidArgumentError, error_msg):
      sampler.generate_random_face_indices(num_samples, face_weights=face_wts)

  @parameterized.parameters(
      ((0., 0.), 10, (5, 5)),
      ((0., 0.0, 0.001), 100, (0, 0, 100)),
      ((0.1, 0.2, 0.3), 1000, (167, 333, 500)),
  )
  def test_random_face_indices(self, face_weights, num_samples, expected):
    """Test for generate_random_face_indices."""
    face_weights = np.array(face_weights, dtype=np.float32)
    expected = np.array(expected, dtype=np.intp)
    sample_faces = sampler.generate_random_face_indices(
        num_samples, face_weights)

    self.assertEqual(sample_faces.shape[0], num_samples)
    self.compare_poisson_equivalence(expected, tf.math.bincount(sample_faces))

# Tests for generate_random_barycentric_coordinates
  @parameterized.parameters(
      ((1,), (tf.int32)),
      ((None,), (tf.int64)),
  )
  def test_random_barycentric_coordinates_exception_not_raised(
      self, shapes, dtypes):
    """Tests that shape exceptions are not raised for random_barycentric_coordinates."""
    self.assert_exception_is_not_raised(
        sampler.generate_random_barycentric_coordinates, shapes, dtypes)

  @parameterized.parameters(
      ("sample_shape must have a rank of 1.", ()),
      ("sample_shape must have a rank of 1.", (4, None)),
  )
  def test_random_barycentric_coordinates_shape_exception_raised(
      self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised for random_barycentric_coordinates."""
    self.assert_exception_is_raised(
        sampler.generate_random_barycentric_coordinates, error_msg, shapes)

  @parameterized.parameters(
      ((5,),),
      ((10, 1, 3),),
  )
  def test_random_barycentric_coordinates(self, sample_shape):
    """Test for generate_random_barycentric_coordinates."""
    sample_shape = np.array(sample_shape, dtype=np.intp)
    random_coordinates = sampler.generate_random_barycentric_coordinates(
        sample_shape=sample_shape)
    coordinate_sum = tf.reduce_sum(input_tensor=random_coordinates, axis=-1)
    expected_coordinate_sum = np.ones(shape=sample_shape)
    self.assertAllClose(expected_coordinate_sum, coordinate_sum)

  # Tests for weighted_random_sample_triangle_mesh
  @parameterized.parameters(
      (((4, 3), (5, 3), (), (5,)),
       (tf.float32, tf.int32, tf.int32, tf.float32)),
      (((None, 3), (None, 3), (), (None,)),
       (tf.float32, tf.int32, tf.int32, tf.float32)),
      (((3, None, 3), (3, None, 3), (), (3, None)),
       (tf.float32, tf.int64, tf.int64, tf.float64)),
      (((3, 6, 5), (3, 5, 3), (), (3, 5)),
       (tf.float64, tf.int32, tf.int32, tf.float32)),
  )
  def test_weighted_sampler_exception_not_raised(self, shapes, dtypes):
    """Tests that the shape exceptions are not raised for weighted sampler."""
    self.assert_exception_is_not_raised(
        sampler.weighted_random_sample_triangle_mesh, shapes, dtypes)

  @parameterized.parameters(
      ("vertex_attributes must have a rank greater than 1.", (3,), (None, 3),
       (), (None, 3)),
      ("faces must have a rank greater than 1.", (5, 2), (None,), (), (None,)),
      ("face_weights must have a rank greater than 0.", (1, None, 3), (None, 3),
       (), ()),
      ("Not all batch dimensions are identical", (4, 4, 2), (3, 5, 3), (),
       (3, 5)),
      ("Not all batch dimensions are identical", (4, 2), (5, 3), (), (4,)),
  )
  def test_weighted_sampler_exception_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised for weighted sampler."""
    self.assert_exception_is_raised(
        sampler.weighted_random_sample_triangle_mesh, error_msg, shapes)

  def test_weighted_sampler_negative_weights(self):
    """Test for exception with negative weights."""
    vertices, faces = mesh_test_utils.create_square_triangle_mesh()
    face_wts = np.array([-0.3, 0.1, 0.5, 0.6], dtype=np.float32)
    num_samples = 10
    error_msg = "Condition x >= y did not hold."
    with self.assertRaisesRegexp(tf.errors.InvalidArgumentError, error_msg):
      sampler.weighted_random_sample_triangle_mesh(
          vertices, faces, num_samples, face_weights=face_wts)

  def test_weighted_random_sample(self):
    """Test for provided face weights."""
    faces = np.array([[0, 1, 2], [2, 1, 3]], dtype=np.int32)
    vertex_attributes = np.array([[0.], [0.], [1.], [1.]], dtype=np.float32)

    # Equal face weights, mean of sampled attributes = 0.5.
    expected_mean = np.array([0.5], dtype=np.float32)
    sample_pts, _ = sampler.weighted_random_sample_triangle_mesh(
        vertex_attributes,
        faces,
        num_samples=1000000,
        face_weights=(0.5, 0.5))
    self.assertAllClose(
        expected_mean,
        tf.reduce_mean(input_tensor=sample_pts, axis=-2),
        atol=1e-3)
    # Face weights biased towards second face, mean > 0.5
    sample_pts, _ = sampler.weighted_random_sample_triangle_mesh(
        vertex_attributes,
        faces,
        num_samples=1000000,
        face_weights=(0.2, 0.8))
    self.assertGreater(
        tf.reduce_mean(input_tensor=sample_pts, axis=-2), expected_mean)

  def test_weighted_sampler_jacobian_random(self):
    """Test the Jacobian of weighted triangle random sampler."""
    tensor_vertex_size = np.random.randint(1, 3)
    tensor_out_shape = np.random.randint(1, 5, size=tensor_vertex_size)
    tensor_out_shape = tensor_out_shape.tolist()
    vertex_axis = np.array(((0., 0., 1), (1., 0., 0.), (0., 1., 0.),
                            (0., 0., -1.), (-1., 0., 0.), (0., -1., 0.)),
                           dtype=np.float32)
    vertex_axis = vertex_axis.reshape([1] * tensor_vertex_size + [6, 3])
    faces = np.array(((0, 1, 2), (0, 2, 4), (0, 4, 5), (0, 5, 1), (3, 2, 1),
                      (3, 4, 2), (3, 5, 4), (3, 1, 5)),
                     dtype=np.int32)
    faces = faces.reshape([1] * tensor_vertex_size + [8, 3])
    index_init = np.tile(faces, tensor_out_shape + [1, 1])
    vertex_scale = np.random.uniform(0.5, 5., tensor_out_shape + [1] * 2)
    vertex_init = vertex_axis * vertex_scale
    index_tensor = tf.convert_to_tensor(value=index_init)
    face_weights = np.random.uniform(
        size=index_init.shape[:index_init.ndim - 1])
    weights_tensor = tf.convert_to_tensor(value=face_weights)

    num_samples = np.random.randint(10, 100)

    def sampler_fn(vertices):
      sample_pts, _ = sampler.weighted_random_sample_triangle_mesh(
          vertices,
          index_tensor,
          num_samples,
          weights_tensor,
          seed=[0, 1],
          stateless=True)
      return sample_pts

    self.assert_jacobian_is_correct_fn(
        sampler_fn, [vertex_init], atol=1e-4, delta=1e-4)

  # Tests for area_weighted_random_sample_triangle_mesh
  @parameterized.parameters(
      (((4, 3), (5, 3), ()), (tf.float32, tf.int32, tf.int32)),
      (((None, 3), (None, 3), ()), (tf.float32, tf.int32, tf.int32)),
      (((3, None, 3), (3, None, 3), ()), (tf.float32, tf.int64, tf.int64)),
      # Test for vertex attributes + positions
      (((3, 6, 5), (3, 5, 3), (), (3, 6, 3)),
       (tf.float64, tf.int32, tf.int32, tf.float32)),
  )
  def test_area_sampler_exception_not_raised(self, shapes, dtypes):
    """Tests that the shape exceptions are not raised for area weighted sampler."""
    self.assert_exception_is_not_raised(
        sampler.area_weighted_random_sample_triangle_mesh, shapes, dtypes)

  @parameterized.parameters(
      ("vertices must have a rank greater than 1.", (3,), (None, 3), ()),
      ("vertices must have greater than 2 dimensions in axis -1.", (5, 2),
       (None, 3), ()),
      ("vertex_positions must have exactly 3 dimensions in axis -1.", (5, 3),
       (None, 3), (), (3, 2)),
  )
  def test_area_sampler_exception_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised for area weighted sampler."""
    self.assert_exception_is_raised(
        sampler.area_weighted_random_sample_triangle_mesh, error_msg, shapes)

  def test_area_sampler_distribution(self):
    """Test for area weighted sampler distribution."""
    vertices, faces = mesh_test_utils.create_single_triangle_mesh()
    vertices = np.repeat(np.expand_dims(vertices, axis=0), 3, axis=0)
    faces = np.repeat(np.expand_dims(faces, axis=0), 3, axis=0)
    num_samples = 5000
    sample_pts, _ = sampler.area_weighted_random_sample_triangle_mesh(
        vertices, faces, num_samples)

    for i in range(3):
      samples = sample_pts[i, ...]
      self.assertEqual(samples.shape[-2], num_samples)
      # Test distribution in 4 quadrants of [0,1]x[0,1]
      v = samples[:, :2] < [0.5, 0.5]
      not_v = tf.logical_not(v)
      quad00 = tf.math.count_nonzero(tf.reduce_all(input_tensor=v, axis=-1))
      quad11 = tf.math.count_nonzero(tf.reduce_all(input_tensor=not_v, axis=-1))
      quad01 = tf.math.count_nonzero(tf.reduce_all(
          input_tensor=tf.stack((v[:, 0], not_v[:, 1]), axis=1), axis=-1))
      quad10 = tf.math.count_nonzero(tf.reduce_all(
          input_tensor=tf.stack((not_v[:, 0], v[:, 1]), axis=1), axis=-1))
      counts = tf.stack((quad00, quad01, quad10, quad11), axis=0)
      expected = np.array(
          [num_samples / 2, num_samples / 4, num_samples / 4, 0],
          dtype=np.float32)
      self.compare_poisson_equivalence(expected, counts)

  def test_face_distribution(self):
    """Test for distribution of face indices with area weighted sampler."""
    vertices, faces = mesh_test_utils.create_square_triangle_mesh()
    num_samples = 1000
    _, sample_faces = sampler.area_weighted_random_sample_triangle_mesh(
        vertices, faces, num_samples)

    # All points should be approx poisson distributed among the 4 faces.
    self.assertEqual(sample_faces.shape[0], num_samples)
    num_faces = faces.shape[0]
    expected = np.array([num_samples / num_faces] * num_faces, dtype=np.intp)
    self.compare_poisson_equivalence(expected, tf.math.bincount(sample_faces))

  def test_area_sampler_jacobian_random(self):
    """Test the Jacobian of area weighted triangle random sampler."""
    tensor_vertex_size = np.random.randint(1, 3)
    tensor_out_shape = np.random.randint(1, 5, size=tensor_vertex_size)
    tensor_out_shape = tensor_out_shape.tolist()
    vertex_axis = np.array(((0., 0., 1), (1., 0., 0.), (0., 1., 0.),
                            (0., 0., -1.), (-1., 0., 0.), (0., -1., 0.)),
                           dtype=np.float32)
    vertex_axis = vertex_axis.reshape([1] * tensor_vertex_size + [6, 3])
    faces = np.array(((0, 1, 2), (0, 2, 4), (0, 4, 5), (0, 5, 1), (3, 2, 1),
                      (3, 4, 2), (3, 5, 4), (3, 1, 5)),
                     dtype=np.int32)
    faces = faces.reshape([1] * tensor_vertex_size + [8, 3])
    index_init = np.tile(faces, tensor_out_shape + [1, 1])
    vertex_scale = np.random.uniform(0.5, 5., tensor_out_shape + [1] * 2)
    vertex_init = vertex_axis * vertex_scale
    index_tensor = tf.convert_to_tensor(value=index_init)

    num_samples = np.random.randint(10, 100)

    def sampler_fn(vertices):
      sample_pts, _ = sampler.area_weighted_random_sample_triangle_mesh(
          vertices, index_tensor, num_samples, seed=[0, 1], stateless=True)
      return sample_pts

    self.assert_jacobian_is_correct_fn(
        sampler_fn, [vertex_init], atol=1e-4, delta=1e-4)


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for 3d rotation matrix."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import flagsaver
from absl.testing import parameterized
import numpy as np
import tensorflow as tf

from tensorflow_graphics.geometry.transformation import axis_angle
from tensorflow_graphics.geometry.transformation import quaternion
from tensorflow_graphics.geometry.transformation import rotation_matrix_3d
from tensorflow_graphics.geometry.transformation.tests import test_data as td
from tensorflow_graphics.geometry.transformation.tests import test_helpers
from tensorflow_graphics.util import test_case


class RotationMatrix3dTest(test_case.TestCase):

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_assert_rotation_matrix_normalized_passthrough(self):
    """Checks that the assert is a passthrough when the flag is False."""
    angles = test_helpers.generate_preset_test_euler_angles()

    matrix_input = rotation_matrix_3d.from_euler(angles)
    matrix_output = rotation_matrix_3d.assert_rotation_matrix_normalized(
        matrix_input)

    self.assertTrue(matrix_input is matrix_output)  # pylint: disable=g-generic-assert

  @parameterized.parameters((np.float32), (np.float64))
  def test_assert_rotation_matrix_normalized_preset(self, dtype):
    """Checks that assert_normalized function works as expected."""
    angles = test_helpers.generate_preset_test_euler_angles().astype(dtype)

    matrix = rotation_matrix_3d.from_euler(angles)
    matrix_rescaled = matrix * 1.01
    matrix_normalized = rotation_matrix_3d.assert_rotation_matrix_normalized(
        matrix)
    self.evaluate(matrix_normalized)

    with self.assertRaises(tf.errors.InvalidArgumentError):  # pylint: disable=g-error-prone-assert-raises
      self.evaluate(rotation_matrix_3d.assert_rotation_matrix_normalized(
          matrix_rescaled))

  @parameterized.parameters(
      ((3, 3),),
      ((None, 3, 3),),
  )
  def test_assert_rotation_matrix_normalized_exception_not_raised(
      self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(
        rotation_matrix_3d.assert_rotation_matrix_normalized, shapes)

  @parameterized.parameters(
      ("must have a rank greater than 1", (3,)),
      ("must have exactly 3 dimensions in axis -1", (3, None)),
      ("must have exactly 3 dimensions in axis -2", (None, 3)),
  )
  def test_assert_rotation_matrix_normalized_exception_raised(
      self, error_msg, *shapes):
    """Tests that the shape exceptions are raised."""
    self.assert_exception_is_raised(
        rotation_matrix_3d.assert_rotation_matrix_normalized, error_msg, shapes)

  @parameterized.parameters(
      ((3,), (1,)),
      ((None, 3), (None, 1)),
      ((1, 3), (1, 1)),
      ((2, 3), (2, 1)),
      ((1, 3), (1,)),
      ((3,), (1, 1)),
  )
  def test_from_axis_angle_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(rotation_matrix_3d.from_axis_angle,
                                        shapes)

  @parameterized.parameters(
      ("must have exactly 3 dimensions in axis -1", (None,), (1,)),
      ("must have exactly 1 dimensions in axis -1", (3,), (None,)),
  )
  def test_from_axis_angle_exception_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(rotation_matrix_3d.from_axis_angle,
                                    error_msg, shapes)

  def test_from_axis_angle_normalized_preset(self):
    """Tests that axis-angles can be converted to rotation matrices."""
    euler_angles = test_helpers.generate_preset_test_euler_angles()

    axis, angle = axis_angle.from_euler(euler_angles)
    matrix_axis_angle = rotation_matrix_3d.from_axis_angle(axis, angle)

    self.assertAllEqual(
        rotation_matrix_3d.is_valid(matrix_axis_angle),
        np.ones(euler_angles.shape[0:-1] + (1,)))

  def test_from_axis_angle_normalized_random(self):
    """Tests that axis-angles can be converted to rotation matrices."""
    tensor_shape = np.random.randint(1, 10, size=np.random.randint(3)).tolist()
    random_axis = np.random.normal(size=tensor_shape + [3])
    random_axis /= np.linalg.norm(random_axis, axis=-1, keepdims=True)
    random_angle = np.random.normal(size=tensor_shape + [1])

    matrix_axis_angle = rotation_matrix_3d.from_axis_angle(
        random_axis, random_angle)

    self.assertAllEqual(
        rotation_matrix_3d.is_valid(matrix_axis_angle),
        np.ones(tensor_shape + [1]))

  @parameterized.parameters(
      ((td.AXIS_3D_X, td.ANGLE_45), (td.MAT_3D_X_45,)),
      ((td.AXIS_3D_Y, td.ANGLE_45), (td.MAT_3D_Y_45,)),
      ((td.AXIS_3D_Z, td.ANGLE_45), (td.MAT_3D_Z_45,)),
      ((td.AXIS_3D_X, td.ANGLE_90), (td.MAT_3D_X_90,)),
      ((td.AXIS_3D_Y, td.ANGLE_90), (td.MAT_3D_Y_90,)),
      ((td.AXIS_3D_Z, td.ANGLE_90), (td.MAT_3D_Z_90,)),
      ((td.AXIS_3D_X, td.ANGLE_180), (td.MAT_3D_X_180,)),
      ((td.AXIS_3D_Y, td.ANGLE_180), (td.MAT_3D_Y_180,)),
      ((td.AXIS_3D_Z, td.ANGLE_180), (td.MAT_3D_Z_180,)),
  )
  def test_from_axis_angle_preset(self, test_inputs, test_outputs):
    """Tests that an axis-angle maps to correct matrix."""
    self.assert_output_is_correct(rotation_matrix_3d.from_axis_angle,
                                  test_inputs, test_outputs)

  def test_from_axis_angle_random(self):
    """Tests conversion to matrix."""
    tensor_shape = np.random.randint(1, 10, size=np.random.randint(3)).tolist()
    random_axis = np.random.normal(size=tensor_shape + [3])
    random_axis /= np.linalg.norm(random_axis, axis=-1, keepdims=True)
    random_angle = np.random.normal(size=tensor_shape + [1])

    matrix_axis_angle = rotation_matrix_3d.from_axis_angle(
        random_axis, random_angle)
    random_quaternion = quaternion.from_axis_angle(random_axis, random_angle)
    matrix_quaternion = rotation_matrix_3d.from_quaternion(random_quaternion)

    self.assertAllClose(matrix_axis_angle, matrix_quaternion, rtol=1e-3)
    # Checks that resulting rotation matrices are normalized.
    self.assertAllEqual(
        rotation_matrix_3d.is_valid(matrix_axis_angle),
        np.ones(tensor_shape + [1]))

  @parameterized.parameters(
      ((td.AXIS_3D_X, td.ANGLE_90, td.AXIS_3D_X), (td.AXIS_3D_X,)),
      ((td.AXIS_3D_X, td.ANGLE_90, td.AXIS_3D_Y), (td.AXIS_3D_Z,)),
      ((td.AXIS_3D_X, -td.ANGLE_90, td.AXIS_3D_Z), (td.AXIS_3D_Y,)),
      ((td.AXIS_3D_Y, -td.ANGLE_90, td.AXIS_3D_X), (td.AXIS_3D_Z,)),
      ((td.AXIS_3D_Y, td.ANGLE_90, td.AXIS_3D_Y), (td.AXIS_3D_Y,)),
      ((td.AXIS_3D_Y, td.ANGLE_90, td.AXIS_3D_Z), (td.AXIS_3D_X,)),
      ((td.AXIS_3D_Z, td.ANGLE_90, td.AXIS_3D_X), (td.AXIS_3D_Y,)),
      ((td.AXIS_3D_Z, -td.ANGLE_90, td.AXIS_3D_Y), (td.AXIS_3D_X,)),
      ((td.AXIS_3D_Z, td.ANGLE_90, td.AXIS_3D_Z), (td.AXIS_3D_Z,)),
  )
  def test_from_axis_angle_rotate_vector_preset(self, test_inputs,
                                                test_outputs):
    """Tests the directionality of axis-angle rotations."""

    def func(axis, angle, point):
      matrix = rotation_matrix_3d.from_axis_angle(axis, angle)
      return rotation_matrix_3d.rotate(point, matrix)

    self.assert_output_is_correct(func, test_inputs, test_outputs)

  @parameterized.parameters(
      ((3,),),
      ((None, 3),),
      ((2, 3),),
  )
  def test_from_euler_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(rotation_matrix_3d.from_euler, shapes)

  @parameterized.parameters(
      ("must have exactly 3 dimensions in axis -1", (None,)),)
  def test_from_euler_exception_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(rotation_matrix_3d.from_euler, error_msg,
                                    shapes)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_from_euler_jacobian_preset(self):
    """Test the Jacobian of the from_euler function."""
    x_init = test_helpers.generate_preset_test_euler_angles()

    self.assert_jacobian_is_correct_fn(rotation_matrix_3d.from_euler, [x_init])

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_from_euler_jacobian_random(self):
    """Test the Jacobian of the from_euler function."""
    x_init = test_helpers.generate_random_test_euler_angles()

    self.assert_jacobian_is_correct_fn(rotation_matrix_3d.from_euler, [x_init])

  def test_from_euler_normalized_preset(self):
    """Tests that euler angles can be converted to rotation matrices."""
    euler_angles = test_helpers.generate_preset_test_euler_angles()

    matrix = rotation_matrix_3d.from_euler(euler_angles)

    self.assertAllEqual(
        rotation_matrix_3d.is_valid(matrix),
        np.ones(euler_angles.shape[0:-1] + (1,)))

  def test_from_euler_normalized_random(self):
    """Tests that euler angles can be converted to rotation matrices."""
    random_euler_angles = test_helpers.generate_random_test_euler_angles()

    matrix = rotation_matrix_3d.from_euler(random_euler_angles)

    self.assertAllEqual(
        rotation_matrix_3d.is_valid(matrix),
        np.ones(random_euler_angles.shape[0:-1] + (1,)))

  @parameterized.parameters(
      ((td.AXIS_3D_0,), (td.MAT_3D_ID,)),
      ((td.ANGLE_45 * td.AXIS_3D_X,), (td.MAT_3D_X_45,)),
      ((td.ANGLE_45 * td.AXIS_3D_Y,), (td.MAT_3D_Y_45,)),
      ((td.ANGLE_45 * td.AXIS_3D_Z,), (td.MAT_3D_Z_45,)),
      ((td.ANGLE_90 * td.AXIS_3D_X,), (td.MAT_3D_X_90,)),
      ((td.ANGLE_90 * td.AXIS_3D_Y,), (td.MAT_3D_Y_90,)),
      ((td.ANGLE_90 * td.AXIS_3D_Z,), (td.MAT_3D_Z_90,)),
      ((td.ANGLE_180 * td.AXIS_3D_X,), (td.MAT_3D_X_180,)),
      ((td.ANGLE_180 * td.AXIS_3D_Y,), (td.MAT_3D_Y_180,)),
      ((td.ANGLE_180 * td.AXIS_3D_Z,), (td.MAT_3D_Z_180,)),
  )
  def test_from_euler_preset(self, test_inputs, test_outputs):
    """Tests that Euler angles create the expected matrix."""
    self.assert_output_is_correct(rotation_matrix_3d.from_euler, test_inputs,
                                  test_outputs)

  def test_from_euler_random(self):
    """Tests that Euler angles produce the same result as axis-angle."""
    angles = test_helpers.generate_random_test_euler_angles()
    matrix = rotation_matrix_3d.from_euler(angles)
    tensor_tile = angles.shape[:-1]

    x_axis = np.tile(td.AXIS_3D_X, tensor_tile + (1,))
    y_axis = np.tile(td.AXIS_3D_Y, tensor_tile + (1,))
    z_axis = np.tile(td.AXIS_3D_Z, tensor_tile + (1,))
    x_angle = np.expand_dims(angles[..., 0], axis=-1)
    y_angle = np.expand_dims(angles[..., 1], axis=-1)
    z_angle = np.expand_dims(angles[..., 2], axis=-1)
    x_rotation = rotation_matrix_3d.from_axis_angle(x_axis, x_angle)
    y_rotation = rotation_matrix_3d.from_axis_angle(y_axis, y_angle)
    z_rotation = rotation_matrix_3d.from_axis_angle(z_axis, z_angle)
    expected_matrix = tf.matmul(z_rotation, tf.matmul(y_rotation, x_rotation))

    self.assertAllClose(expected_matrix, matrix, rtol=1e-3)

  @parameterized.parameters(
      ((3,),),
      ((None, 3),),
  )
  def test_from_euler_with_small_angles_approximation_exception_not_raised(
      self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(
        rotation_matrix_3d.from_euler_with_small_angles_approximation, shapes)

  @parameterized.parameters(
      ("must have exactly 3 dimensions in axis -1", (None,)),)
  def test_from_euler_with_small_angles_approximation_exception_raised(
      self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(
        rotation_matrix_3d.from_euler_with_small_angles_approximation,
        error_msg, shapes)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_from_euler_with_small_angles_approximation_jacobian_random(self):
    """Test the Jacobian of from_euler_with_small_angles_approximation."""
    x_init = test_helpers.generate_random_test_euler_angles(
        min_angle=-0.17, max_angle=0.17)

    self.assert_jacobian_is_correct_fn(
        rotation_matrix_3d.from_euler_with_small_angles_approximation, [x_init])

  def test_from_euler_with_small_angles_approximation_random(self):
    """Tests small_angles approximation by comparing to exact calculation."""
    # Only generate small angles. For a test tolerance of 1e-3, 0.16 was found
    # empirically to be the range where the small angle approximation works.
    random_euler_angles = test_helpers.generate_random_test_euler_angles(
        min_angle=-0.16, max_angle=0.16)

    exact_matrix = rotation_matrix_3d.from_euler(random_euler_angles)
    approximate_matrix = (
        rotation_matrix_3d.from_euler_with_small_angles_approximation(
            random_euler_angles))

    self.assertAllClose(exact_matrix, approximate_matrix, atol=1e-3)

  @parameterized.parameters(
      ((4,),),
      ((None, 4),),
  )
  def test_from_quaternion_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(rotation_matrix_3d.from_quaternion,
                                        shapes)

  @parameterized.parameters(
      ("must have exactly 4 dimensions in axis -1", (None,)),)
  def test_from_quaternion_exception_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(rotation_matrix_3d.from_quaternion,
                                    error_msg, shapes)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_from_quaternion_jacobian_preset(self):
    """Test the Jacobian of the from_quaternion function."""
    x_init = test_helpers.generate_preset_test_quaternions()

    self.assert_jacobian_is_correct_fn(rotation_matrix_3d.from_quaternion,
                                       [x_init])

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_from_quaternion_jacobian_random(self):
    """Test the Jacobian of the from_quaternion function."""
    x_init = test_helpers.generate_random_test_quaternions()

    self.assert_jacobian_is_correct_fn(rotation_matrix_3d.from_quaternion,
                                       [x_init])

  def test_from_quaternion_normalized_preset(self):
    """Tests that quaternions can be converted to rotation matrices."""
    euler_angles = test_helpers.generate_preset_test_euler_angles()

    quat = quaternion.from_euler(euler_angles)
    matrix_quat = rotation_matrix_3d.from_quaternion(quat)

    self.assertAllEqual(
        rotation_matrix_3d.is_valid(matrix_quat),
        np.ones(euler_angles.shape[0:-1] + (1,)))

  def test_from_quaternion_normalized_random(self):
    """Tests that random quaternions can be converted to rotation matrices."""
    random_quaternion = test_helpers.generate_random_test_quaternions()
    tensor_shape = random_quaternion.shape[:-1]

    random_matrix = rotation_matrix_3d.from_quaternion(random_quaternion)

    self.assertAllEqual(
        rotation_matrix_3d.is_valid(random_matrix),
        np.ones(tensor_shape + (1,)))

  def test_from_quaternion_preset(self):
    """Tests that a quaternion maps to correct matrix."""
    preset_quaternions = test_helpers.generate_preset_test_quaternions()

    preset_matrices = test_helpers.generate_preset_test_rotation_matrices_3d()

    self.assertAllClose(preset_matrices,
                        rotation_matrix_3d.from_quaternion(preset_quaternions))

  def test_from_quaternion_random(self):
    """Tests conversion to matrix."""
    random_euler_angles = test_helpers.generate_random_test_euler_angles()

    random_quaternions = quaternion.from_euler(random_euler_angles)
    random_rotation_matrices = rotation_matrix_3d.from_euler(
        random_euler_angles)

    self.assertAllClose(random_rotation_matrices,
                        rotation_matrix_3d.from_quaternion(random_quaternions))

  @parameterized.parameters(
      ((3, 3),),
      ((None, 3, 3),),
      ((2, 3, 3),),
  )
  def test_inverse_exception_not_raised(self, *shapes):
    """Checks the inputs of the rotate function."""
    self.assert_exception_is_not_raised(rotation_matrix_3d.inverse, shapes)

  @parameterized.parameters(
      ("must have a rank greater than 1", (3,)),
      ("must have exactly 3 dimensions in axis -1", (3, None)),
      ("must have exactly 3 dimensions in axis -2", (None, 3)),
  )
  def test_inverse_exception_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(rotation_matrix_3d.inverse, error_msg,
                                    shapes)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_inverse_jacobian_preset(self):
    """Test the Jacobian of the inverse function."""
    x_init = test_helpers.generate_preset_test_rotation_matrices_3d()

    self.assert_jacobian_is_correct_fn(rotation_matrix_3d.inverse, [x_init])

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_inverse_jacobian_random(self):
    """Test the Jacobian of the inverse function."""
    x_init = test_helpers.generate_random_test_rotation_matrix_3d()

    self.assert_jacobian_is_correct_fn(rotation_matrix_3d.inverse, [x_init])

  def test_inverse_normalized_random(self):
    """Checks that inverted rotation matrices are valid rotations."""
    random_euler_angle = test_helpers.generate_random_test_euler_angles()
    tensor_tile = random_euler_angle.shape[:-1]

    random_matrix = rotation_matrix_3d.from_euler(random_euler_angle)
    predicted_invert_random_matrix = rotation_matrix_3d.inverse(random_matrix)

    self.assertAllEqual(
        rotation_matrix_3d.is_valid(predicted_invert_random_matrix),
        np.ones(tensor_tile + (1,)))

  def test_inverse_random(self):
    """Checks that inverting rotated points results in no transformation."""
    random_euler_angle = test_helpers.generate_random_test_euler_angles()
    tensor_tile = random_euler_angle.shape[:-1]
    random_matrix = rotation_matrix_3d.from_euler(random_euler_angle)
    random_point = np.random.normal(size=tensor_tile + (3,))

    rotated_random_points = rotation_matrix_3d.rotate(random_point,
                                                      random_matrix)
    predicted_invert_random_matrix = rotation_matrix_3d.inverse(random_matrix)
    predicted_invert_rotated_random_points = rotation_matrix_3d.rotate(
        rotated_random_points, predicted_invert_random_matrix)

    self.assertAllClose(
        random_point, predicted_invert_rotated_random_points, rtol=1e-6)

  @parameterized.parameters(
      ((3, 3),),
      ((None, 3, 3),),
      ((2, 3, 3),),
  )
  def test_is_valid_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(rotation_matrix_3d.is_valid, shapes)

  @parameterized.parameters(
      ("must have a rank greater than 1", (3,)),
      ("must have exactly 3 dimensions in axis -1", (3, None)),
      ("must have exactly 3 dimensions in axis -2", (None, 3)),
  )
  def test_is_valid_exception_raised(self, error_msg, *shape):
    """Tests that the shape exceptions are raised."""
    self.assert_exception_is_raised(rotation_matrix_3d.is_valid, error_msg,
                                    shape)

  def test_is_valid_random(self):
    """Tests that is_valid works as intended."""
    random_euler_angle = test_helpers.generate_random_test_euler_angles()
    tensor_tile = random_euler_angle.shape[:-1]

    rotation_matrix = rotation_matrix_3d.from_euler(random_euler_angle)
    pred_normalized = rotation_matrix_3d.is_valid(rotation_matrix)

    with self.subTest(name="all_normalized"):
      self.assertAllEqual(pred_normalized,
                          np.ones(shape=tensor_tile + (1,), dtype=bool))

    with self.subTest(name="non_orthonormal"):
      test_matrix = np.array([[2., 0., 0.], [0., 0.5, 0], [0., 0., 1.]])
      pred_normalized = rotation_matrix_3d.is_valid(test_matrix)

      self.assertAllEqual(pred_normalized, np.zeros(shape=(1,), dtype=bool))

    with self.subTest(name="negative_orthonormal"):
      test_matrix = np.array([[1., 0., 0.], [0., -1., 0.], [0., 0., 1.]])
      pred_normalized = rotation_matrix_3d.is_valid(test_matrix)

      self.assertAllEqual(pred_normalized, np.zeros(shape=(1,), dtype=bool))

  @parameterized.parameters(
      ((3,), (3, 3)),
      ((None, 3), (None, 3, 3)),
      ((1, 3), (1, 3, 3)),
      ((2, 3), (2, 3, 3)),
      ((3,), (1, 3, 3)),
      ((1, 3), (3, 3)),
  )
  def test_rotate_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(rotation_matrix_3d.rotate, shapes)

  @parameterized.parameters(
      ("must have exactly 3 dimensions in axis -1", (None,), (3, 3)),
      ("must have a rank greater than 1", (3,), (3,)),
      ("must have exactly 3 dimensions in axis -1", (3,), (3, None)),
      ("must have exactly 3 dimensions in axis -2", (3,), (None, 3)),
  )
  def test_rotate_exception_raised(self, error_msg, *shapes):
    """Checks the inputs of the rotate function."""
    self.assert_exception_is_raised(rotation_matrix_3d.rotate, error_msg,
                                    shapes)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_rotate_jacobian_preset(self):
    """Test the Jacobian of the rotate function."""
    x_matrix_init = test_helpers.generate_preset_test_rotation_matrices_3d()
    tensor_shape = x_matrix_init.shape[:-1]
    x_point_init = np.random.uniform(size=tensor_shape)

    self.assert_jacobian_is_correct_fn(rotation_matrix_3d.rotate,
                                       [x_point_init, x_matrix_init])

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_rotate_jacobian_random(self):
    """Test the Jacobian of the rotate function."""
    x_matrix_init = test_helpers.generate_random_test_rotation_matrix_3d()
    tensor_shape = x_matrix_init.shape[:-1]
    x_point_init = np.random.uniform(size=tensor_shape)

    self.assert_jacobian_is_correct_fn(rotation_matrix_3d.rotate,
                                       [x_point_init, x_matrix_init])

  @parameterized.parameters(
      ((td.ANGLE_90 * td.AXIS_3D_X, td.AXIS_3D_X), (td.AXIS_3D_X,)),
      ((td.ANGLE_90 * td.AXIS_3D_X, td.AXIS_3D_Y), (td.AXIS_3D_Z,)),
      ((-td.ANGLE_90 * td.AXIS_3D_X, td.AXIS_3D_Z), (td.AXIS_3D_Y,)),
      ((-td.ANGLE_90 * td.AXIS_3D_Y, td.AXIS_3D_X), (td.AXIS_3D_Z,)),
      ((td.ANGLE_90 * td.AXIS_3D_Y, td.AXIS_3D_Y), (td.AXIS_3D_Y,)),
      ((td.ANGLE_90 * td.AXIS_3D_Y, td.AXIS_3D_Z), (td.AXIS_3D_X,)),
      ((td.ANGLE_90 * td.AXIS_3D_Z, td.AXIS_3D_X), (td.AXIS_3D_Y,)),
      ((-td.ANGLE_90 * td.AXIS_3D_Z, td.AXIS_3D_Y), (td.AXIS_3D_X,)),
      ((td.ANGLE_90 * td.AXIS_3D_Z, td.AXIS_3D_Z), (td.AXIS_3D_Z,)),
  )
  def test_rotate_vector_preset(self, test_inputs, test_outputs):
    """Tests that the rotate function produces the expected results."""

    def func(angles, point):
      matrix = rotation_matrix_3d.from_euler(angles)
      return rotation_matrix_3d.rotate(point, matrix)

    self.assert_output_is_correct(func, test_inputs, test_outputs)

  def test_rotate_vs_rotate_quaternion_random(self):
    """Tests that the rotate provide the same results as quaternion.rotate."""
    random_euler_angle = test_helpers.generate_random_test_euler_angles()
    tensor_tile = random_euler_angle.shape[:-1]

    random_matrix = rotation_matrix_3d.from_euler(random_euler_angle)
    random_quaternion = quaternion.from_rotation_matrix(random_matrix)
    random_point = np.random.normal(size=tensor_tile + (3,))
    ground_truth = quaternion.rotate(random_point, random_quaternion)
    prediction = rotation_matrix_3d.rotate(random_point, random_matrix)

    self.assertAllClose(ground_truth, prediction, rtol=1e-6)


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module implements the rendering equation for a point light."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math
import tensorflow as tf

from tensorflow_graphics.math import vector
from tensorflow_graphics.util import asserts
from tensorflow_graphics.util import export_api
from tensorflow_graphics.util import shape


def estimate_radiance(point_light_radiance,
                      point_light_position,
                      surface_point_position,
                      surface_point_normal,
                      observation_point,
                      brdf,
                      name=None,
                      reflected_light_fall_off=False):
  """Estimates the spectral radiance of a point light reflected from the surface point towards the observation point.

  Note:
    In the following, A1 to An are optional batch dimensions, which must be
    broadcast compatible.
    B1 to Bm are optional batch dimensions for the lights, which must be
    broadcast compatible.

  Note:
    In case the light or the observation point are located behind the surface
    the function will return 0.

  Note:
    The gradient of this function is not smooth when the dot product of the
    normal with the light-to-surface or surface-to-observation vectors is 0.

  Args:
    point_light_radiance: A tensor of shape '[B1, ..., Bm, K]', where the last
      axis represents the radiance of the point light at a specific wave length.
    point_light_position: A tensor of shape `[B1, ..., Bm, 3]`, where the last
      axis represents the position of the point light.
    surface_point_position: A tensor of shape `[A1, ..., An, 3]`, where the last
      axis represents the position of the surface point.
    surface_point_normal: A tensor of shape `[A1, ..., An, 3]`, where the last
      axis represents the normalized surface normal at the given surface point.
    observation_point: A tensor of shape `[A1, ..., An, 3]`, where the last axis
      represents the observation point.
    brdf: The BRDF of the surface as a function of:
      incoming_light_direction - The incoming light direction as the last axis
      of a tensor with shape `[A1, ..., An, 3]`.
      outgoing_light_direction - The outgoing light direction as the last axis
      of a tensor with shape `[A1, ..., An, 3]`.
      surface_point_normal - The surface normal as the last axis of a tensor
      with shape `[A1, ..., An, 3]`.
      Note - The BRDF should return a tensor of size '[A1, ..., An, K]' where
      the last axis represents the amount of reflected light in each wave
      length.
    name: A name for this op. Defaults to "estimate_radiance".
    reflected_light_fall_off: A boolean specifying whether or not to include the
      fall off of the light reflected from the surface towards the observation
      point in the calculation. Defaults to False.

  Returns:
    A tensor of shape `[A1, ..., An, B1, ..., Bm, K]`, where the last
      axis represents the amount of light received at the observation point
      after being reflected from the given surface point.

  Raises:
    ValueError: if the shape of `point_light_position`,
    `surface_point_position`, `surface_point_normal`, or `observation_point` is
    not supported.
    InvalidArgumentError: if 'surface_point_normal' is not normalized.
  """
  with tf.compat.v1.name_scope(name, "estimate_radiance", [
      point_light_radiance, point_light_position, surface_point_position,
      surface_point_normal, observation_point, brdf
  ]):
    point_light_radiance = tf.convert_to_tensor(value=point_light_radiance)
    point_light_position = tf.convert_to_tensor(value=point_light_position)
    surface_point_position = tf.convert_to_tensor(value=surface_point_position)
    surface_point_normal = tf.convert_to_tensor(value=surface_point_normal)
    observation_point = tf.convert_to_tensor(value=observation_point)

    shape.check_static(
        tensor=point_light_position,
        tensor_name="point_light_position",
        has_dim_equals=(-1, 3))
    shape.check_static(
        tensor=surface_point_position,
        tensor_name="surface_point_position",
        has_dim_equals=(-1, 3))
    shape.check_static(
        tensor=surface_point_normal,
        tensor_name="surface_point_normal",
        has_dim_equals=(-1, 3))
    shape.check_static(
        tensor=observation_point,
        tensor_name="observation_point",
        has_dim_equals=(-1, 3))
    shape.compare_batch_dimensions(
        tensors=(surface_point_position, surface_point_normal,
                 observation_point),
        tensor_names=("surface_point_position", "surface_point_normal",
                      "observation_point"),
        last_axes=-2,
        broadcast_compatible=True)
    shape.compare_batch_dimensions(
        tensors=(point_light_radiance, point_light_position),
        tensor_names=("point_light_radiance", "point_light_position"),
        last_axes=-2,
        broadcast_compatible=True)
    surface_point_normal = asserts.assert_normalized(surface_point_normal)

    # Get the number of lights dimensions (B1,...,Bm).
    lights_num_dimensions = max(
        len(point_light_radiance.shape), len(point_light_position.shape)) - 1
    # Reshape the other parameters so they can be broadcasted to the output of
    # shape [A1,...,An, B1,...,Bm, K].
    surface_point_position = tf.reshape(
        surface_point_position,
        surface_point_position.shape[:-1] + (1,) * lights_num_dimensions + (3,))
    surface_point_normal = tf.reshape(
        surface_point_normal,
        surface_point_normal.shape[:-1] + (1,) * lights_num_dimensions + (3,))
    observation_point = tf.reshape(
        observation_point,
        observation_point.shape[:-1] + (1,) * lights_num_dimensions + (3,))

    light_to_surface_point = surface_point_position - point_light_position
    distance_light_surface_point = tf.norm(
        tensor=light_to_surface_point, axis=-1, keepdims=True)
    incoming_light_direction = tf.math.l2_normalize(
        light_to_surface_point, axis=-1)
    surface_to_observation_point = observation_point - surface_point_position
    outgoing_light_direction = tf.math.l2_normalize(
        surface_to_observation_point, axis=-1)
    brdf_value = brdf(incoming_light_direction, outgoing_light_direction,
                      surface_point_normal)
    incoming_light_dot_surface_normal = vector.dot(-incoming_light_direction,
                                                   surface_point_normal)
    outgoing_light_dot_surface_normal = vector.dot(outgoing_light_direction,
                                                   surface_point_normal)

    estimated_radiance = (point_light_radiance * \
                          brdf_value * incoming_light_dot_surface_normal) / \
        (4. * math.pi * tf.math.square(distance_light_surface_point))

    if reflected_light_fall_off:
      distance_surface_observation_point = tf.norm(
          tensor=surface_to_observation_point, axis=-1, keepdims=True)
      estimated_radiance = estimated_radiance / \
          tf.math.square(distance_surface_observation_point)

    # Create a condition for checking whether the light or observation point are
    # behind the surface.
    min_dot = tf.minimum(incoming_light_dot_surface_normal,
                         outgoing_light_dot_surface_normal)
    common_shape = shape.get_broadcasted_shape(min_dot.shape,
                                               estimated_radiance.shape)
    d_val = lambda dim: 1 if dim is None else tf.compat.v1.dimension_value(dim)
    common_shape = [d_val(dim) for dim in common_shape]
    condition = tf.broadcast_to(tf.greater_equal(min_dot, 0.0), common_shape)

    return tf.compat.v1.where(condition, estimated_radiance,
                              tf.zeros_like(estimated_radiance))


# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()
